/*
 *        spAudio_AU.c: Audio Unit on Mac OS X
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spThread.h>

#include <sp/spAudioP.h>

#if defined(SP_SUPPORT_AUDIO) && defined(MACOSX) && defined(SP_MAC_USE_AUDIO_UNIT)
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#if defined(IPHONE)
#include <AudioToolbox/AudioToolbox.h>
#else
#include <CoreAudio/AudioHardware.h>
#endif

#define kOutputBus 0
#define kInputBus 1

#define SP_AUDIO_AU_BUFFER_MARGIN /*32*/0
#if 0 && defined(IPHONE)
#define SP_AUDIO_AU_MIN_BUFFER_SIZE 8192
#else
#define SP_AUDIO_AU_MIN_BUFFER_SIZE /*2048*//*1024*/32
#endif

#define SP_AUDIO_AU_AUDIO_BUFFER_LIST_BUFFER_SIZE_MIN_INCREMENT 32

#if defined(IPHONE) && defined(CA_CANONICAL_DEPRECATED)
#if defined(CA_PREFER_FIXED_POINT) && CA_PREFER_FIXED_POINT
#define spAudioUnitSampleType SInt32
#define SP_SYSTEM_AUDIO_UNIT_FORMAT_FLAGS (kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved)
#else /* defined(CA_PREFER_FIXED_POINT) && CA_PREFER_FIXED_POINT */
#define spAudioUnitSampleType Float32
#define SP_SYSTEM_AUDIO_UNIT_FORMAT_FLAGS (kAudioFormatFlagIsFloat | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved)
#endif /* !(defined(CA_PREFER_FIXED_POINT) && CA_PREFER_FIXED_POINT) */
#else /* defined(IPHONE) && defined(CA_CANONICAL_DEPRECATED) */
typedef Float32 spAudioUnitSampleType;
#if !defined(IPHONE)
/*typedef SInt32 AudioUnitSampleType;*/
#define SP_SYSTEM_AUDIO_UNIT_FORMAT_FLAGS (kAudioFormatFlagIsFloat | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved)
#else
#define SP_SYSTEM_AUDIO_UNIT_FORMAT_FLAGS kAudioFormatFlagsAudioUnitCanonical
#endif
#endif /* !(defined(IPHONE) && defined(CA_CANONICAL_DEPRECATED)) */

static void updateFormatInformation(spAudio audio)
{
    UInt32 sampByte;

    sampByte = spGetBitByte(audio->samp_bit, NULL);
    sampByte = (audio->samp_bit <= 32 && sampByte >= 8) ? 4 : sampByte;
    spDebug(50, "updateFormatInformation", "sampByte = %d\n", sampByte);
    audio->currentFormat.mBitsPerChannel = 8 * sampByte;

    if (audio->currentFormat.mFormatFlags & kAudioFormatFlagIsNonInterleaved) {
        audio->currentFormat.mBytesPerFrame = sampByte;
    } else {
        audio->currentFormat.mBytesPerFrame = sampByte * audio->currentFormat.mChannelsPerFrame;
    }
    audio->currentFormat.mBytesPerPacket = audio->currentFormat.mBytesPerFrame;
    
    sampByte = audio->systemFormat.mBitsPerChannel / 8;

    if (audio->systemFormat.mFormatFlags & kAudioFormatFlagIsNonInterleaved) {
        audio->systemFormat.mBytesPerFrame = sampByte;
    } else {
        audio->systemFormat.mBytesPerFrame = sampByte * audio->systemFormat.mChannelsPerFrame;
    }
    audio->systemFormat.mBytesPerPacket = audio->systemFormat.mBytesPerFrame;

    return;
}

static void resetBufferCounter(spAudio audio)
{
    AudioTimeStamp timeStamp;
    
    audio->inputSampleTime = audio->outputSampleTime = 0.0;
    audio->inputInitialSampleTime = audio->outputInitialSampleTime = -1.0;
    audio->prevCallbackedOutputPosition = 0;
    
    audio->inputBufferOffset = audio->outputBufferOffset = 0;
    audio->inputBufferFilled = audio->outputBufferFilled = 0;
    audio->inputBufferReadPos = audio->outputBufferWritePos = 0;
    audio->inputCallbackCount = audio->outputCallbackCount = 0;
            
    audio->inputConvertedFrames = audio->inputConvertedOffset = 0;
    audio->outputConvertedFrames = audio->outputConvertedOffset = 0;
            
#if !defined(IPHONE) || __IPHONE_OS_VERSION_MAX_ALLOWED >= 60000
    if (audio->inputEnabled == SP_TRUE && audio->inputUnitStarted == SP_TRUE) {
        memset(&timeStamp, 0, sizeof(AudioTimeStamp));
        timeStamp.mFlags = kAudioTimeStampSampleTimeValid;
        timeStamp.mSampleTime = -1.0;
        
        if (AudioUnitSetProperty(audio->inputUnit, kAudioUnitProperty_ScheduleStartTimeStamp, 
                                 kAudioUnitScope_Global, kInputBus, &timeStamp, sizeof(timeStamp)) != noErr) {
            spDebug(10, "resetBufferCounter", "AudioUnitSetProperty of kAudioUnitProperty_ScheduleStartTimeStamp for input\n");
        }
    }
    
    if (audio->outputEnabled == SP_TRUE && audio->outputUnitStarted == SP_TRUE) {
        memset(&timeStamp, 0, sizeof(AudioTimeStamp));
        timeStamp.mFlags = kAudioTimeStampSampleTimeValid;
        timeStamp.mSampleTime = -1.0;
        
        if (AudioUnitSetProperty(audio->outputUnit, kAudioUnitProperty_ScheduleStartTimeStamp, 
                                 kAudioUnitScope_Global, kOutputBus, &timeStamp, sizeof(timeStamp)) != noErr) {
            spDebug(10, "resetBufferCounter", "AudioUnitSetProperty of kAudioUnitProperty_ScheduleStartTimeStamp for output\n");
        }
    }
#endif
    
    audio->inputWaiting = SP_FALSE;
    audio->outputWaiting = SP_FALSE;
    
    return;
}

static spBool allocInternalBuffer(spAudio audio)
{
    UInt32 bufferSize;
    
    audio->inputBufferSize = 0;
    audio->inputBuffer = NULL;
    
    if (audio->inputEnabled == SP_TRUE) {
        bufferSize = audio->bufferSizeForInput * MAX(audio->num_buffer, 2);
        if ((audio->inputBuffer = xalloc(bufferSize * audio->currentFormat.mChannelsPerFrame, char)) == NULL) {
            return SP_FALSE;
        }
        memset(audio->inputBuffer, 0, bufferSize * audio->currentFormat.mChannelsPerFrame);
        audio->inputBufferSize = bufferSize;
    }
    
    audio->outputBufferSize = 0;
    audio->outputBuffer = NULL;
    
    if (audio->outputEnabled == SP_TRUE) {
        bufferSize = audio->bufferSizeForOutput * MAX(audio->num_buffer, 2);
        if ((audio->outputBuffer = xalloc(bufferSize * audio->currentFormat.mChannelsPerFrame, char)) == NULL) {
            return SP_FALSE;
        }
        memset(audio->outputBuffer, 0, bufferSize * audio->currentFormat.mChannelsPerFrame);
        audio->outputBufferSize = bufferSize;
    }
    
    spDebug(10, "allocInternalBuffer", "inputBufferSize = %d, outputBufferSize = %d\n",
            audio->inputBufferSize, audio->outputBufferSize);
    
    return SP_TRUE;
}

static void freeInternalBuffer(spAudio audio, spBool outputFlag)
{
    spDebug(80, "freeInternalBuffer", "in\n");

    if (outputFlag == SP_FALSE) {
        if (audio->inputBuffer != NULL) {
            xfree(audio->inputBuffer);
            audio->inputBuffer = NULL;
        }
        audio->inputBufferSize = 0;
    } else {
        if (audio->outputBuffer != NULL) {
            xfree(audio->outputBuffer);
            audio->outputBuffer = NULL;
        }
        audio->outputBufferSize = 0;
    }
    
    spDebug(80, "freeInternalBuffer", "done\n");
    
    return;
}

static spBool allocConversionBuffer(spAudio audio)
{
    UInt32 bufferSize;
    
    if (audio->inputEnabled == SP_TRUE) {
        bufferSize = MAX(2 * audio->currentFormat.mChannelsPerFrame * audio->bufferSizeForInput, SP_AUDIO_AU_MIN_BUFFER_SIZE);
        
        if ((audio->inputConversionBuffer = xalloc(bufferSize + SP_AUDIO_AU_BUFFER_MARGIN, char)) == NULL) {
            return SP_FALSE;
        }
    } else {
        audio->inputConversionBuffer = NULL;
    }
    if (audio->outputEnabled == SP_TRUE) {
        bufferSize = MAX(2 * audio->currentFormat.mChannelsPerFrame * audio->bufferSizeForOutput, SP_AUDIO_AU_MIN_BUFFER_SIZE);
        
        if ((audio->outputConversionBuffer = xalloc(bufferSize + SP_AUDIO_AU_BUFFER_MARGIN, char)) == NULL) {
            return SP_FALSE;
        }
    } else {
        audio->outputConversionBuffer = NULL;
    }

    return SP_TRUE;
}

static void freeConversionBuffer(spAudio audio, spBool outputFlag)
{
    spDebug(80, "freeConversionBuffer", "in: outputFlag = %d\n", outputFlag);

    if (outputFlag == SP_FALSE) {
        if (audio->inputConversionBuffer != NULL) {
            xfree(audio->inputConversionBuffer);
            audio->inputConversionBuffer = NULL;
        }
    } else {
        if (audio->outputConversionBuffer != NULL) {
            xfree(audio->outputConversionBuffer);
            audio->outputConversionBuffer = NULL;
        }
    }
    
    spDebug(80, "freeConversionBuffer", "done\n");
    
    return;
}

static void freeAudioBufferList(AudioBufferList* list)
{
    UInt32 i;
    
    spDebug(80, "freeAudioBufferList", "in\n");
    
    if (list != NULL) {
        spDebug(80, "freeAudioBufferList", "list->mNumberBuffers = %d\n", list->mNumberBuffers);
        for (i = 0; i < list->mNumberBuffers; i++) {
            if (list->mBuffers[i].mData != NULL) {
                free(list->mBuffers[i].mData);
            }
        }
        free(list);
    }
    
    spDebug(80, "freeAudioBufferList", "done\n");

    return;
}

static AudioBufferList *allocAudioBufferList(UInt32 numChannels, UInt32 size)
{
    UInt32 i;
    AudioBufferList *list;

    list = calloc(1, sizeof(AudioBufferList) + (numChannels - 1) * sizeof(AudioBuffer));
    if (list == NULL) {
        return NULL;
    }
        
    list->mNumberBuffers = numChannels;
    for (i = 0; i < list->mNumberBuffers; i++) {
        list->mBuffers[i].mNumberChannels = 1;
        list->mBuffers[i].mDataByteSize = size;
        if (size > 0) {
            list->mBuffers[i].mData = malloc(size);
            if (list->mBuffers[i].mData == NULL) {
                freeAudioBufferList(list);
                return NULL;
            }
        } else {
            list->mBuffers[i].mData = NULL;
        }
        spDebug(100, "allocAudioBufferList", "list->mBuffers[%d].mData = %ld\n", i, (long)list->mBuffers[i].mData);
    }
    
    spDebug(100, "allocAudioBufferList", "done: size = %d, list = %ld\n", size, (long)list);
    
    return list;
}

static void updateAudioBufferList(AudioBufferList* list, UInt32 *allocSize, UInt32 newSize, spBool doClear)
{
    UInt32 i;
    
    spDebug(80, "updateAudioBufferList", "newSize = %d, allocSize = %d\n", newSize, *allocSize);
    
    if (newSize > *allocSize) {
        *allocSize = MAX(*allocSize + SP_AUDIO_AU_AUDIO_BUFFER_LIST_BUFFER_SIZE_MIN_INCREMENT, newSize);
        spDebug(80, "updateAudioBufferList", "allocSize updated: %d\n", *allocSize);
    }
    
    for (i = 0; i < list->mNumberBuffers; i++) {
        if (newSize != list->mBuffers[i].mDataByteSize) {
            if (newSize > list->mBuffers[i].mDataByteSize) {
                list->mBuffers[i].mData = realloc(list->mBuffers[i].mData, *allocSize);
            }
            list->mBuffers[i].mDataByteSize = newSize;
        }
        if (doClear) {
            memset(list->mBuffers[i].mData, 0, newSize);
        }
        spDebug(100, "updateAudioBufferList", "list->mBuffers[%d].mData = %ld\n",
                i, (long)list->mBuffers[i].mData);
    }

    return;
}

static void clearAudioBufferList(AudioBufferList* list)
{
    UInt32 i;
    
    spDebug(100, "clearAudioBufferList", "list->mNumberBuffers = %d\n", list->mNumberBuffers);
    
    for (i = 0; i < list->mNumberBuffers; i++) {
        spDebug(100, "clearAudioBufferList", "list->mBuffers[%d].mDataByteSize = %d, list->mBuffers[%d].mData = %ld\n",
                i, list->mBuffers[i].mDataByteSize, i, (long)list->mBuffers[i].mData);
        if (list->mBuffers[i].mData != NULL && list->mBuffers[i].mDataByteSize > 0) {
            memset(list->mBuffers[i].mData, 0, list->mBuffers[i].mDataByteSize);
        }
    }
    
    spDebug(100, "clearAudioBufferList", "done\n");
    
    return;
}

static spBool emitOutputPositionCallback(spAudio audio)
{
    spBool noerr = SP_TRUE;

    spDebug(100, "emitOutputPositionCallback", "audio->call_type = %d\n", audio->call_type);
    
    if (audio->call_func != NULL && audio->doStop == SP_FALSE) {
        if (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK) {
            spLong position;

            if (spGetAudioOutputPosition(audio, &position) == SP_TRUE
                && position > audio->prevCallbackedOutputPosition) {
                spDebug(100, "emitOutputPositionCallback", "position = %ld\n", (long)position);
                audio->prevCallbackedOutputPosition = position;
                if (audio->call_func(audio, SP_AUDIO_OUTPUT_POSITION_CALLBACK,
                                     &position, NULL, audio->call_data) == SP_FALSE) {
                    noerr = SP_FALSE;
                }
            }
        }
    }

    return noerr;
}

static void waitOutputAudio(spAudio audio)
{
    spDebug(80, "waitOutputAudio", "spLockMutex...\n");
    
    spLockMutex(audio->outputMutex);
    audio->outputWaiting = SP_TRUE;
    spUnlockMutex(audio->outputMutex);

    spDebug(80, "waitOutputAudio", "spUnlockMutex done\n");
    
    while (audio->outputBufferFilled > 0) {
        spDebug(100, "waitOutputAudio", "audio->outputBufferFilled = %d\n", audio->outputBufferFilled);
        spResetEvent(audio->outputEvent);
        spDebug(80, "waitOutputAudio", "spWaitEvent...\n");
        spWaitEvent(audio->outputEvent);
        spDebug(80, "waitOutputAudio", "spWaitEvent done\n");
        
        if (emitOutputPositionCallback(audio) == SP_FALSE) {
            break;
        }
    }

    audio->outputWaiting = SP_FALSE;
    
    spDebug(10, "waitOutputAudio", "done\n");
    
    return;
}

int spGetNumAudioDriverArch(void)
{
    return 1;
}

const char *spGetAudioDriverNameArch(int index)
{
    if (index == 0) {
        return "Default Device";
    }
    
    return NULL;
}

static OSStatus inputConverterCallback(AudioConverterRef inAudioConverter, 
                                       UInt32 *ioNumberDataPackets, 
                                       AudioBufferList *ioData, 
                                       AudioStreamPacketDescription **outDataPacketDescription, 
                                       void *inUserData)
{
    UInt32 i;
    UInt32 bytesCopied = 0;
    spAudio audio = (spAudio)inUserData;

    if (*ioNumberDataPackets > 0) {
        bytesCopied = *ioNumberDataPackets * audio->systemFormat.mBytesPerPacket;
        spDebug(80, "inputConverterCallback",
                "*ioNumberDataPackets = %d, bytesCopied = %d, inputDeviceBufferList->mBuffers[0].mDataByteSize = %d, inputDeviceBufferList->mBuffers[0].mNumberChannels = %d\n",
                *ioNumberDataPackets, bytesCopied, audio->inputDeviceBufferList->mBuffers[0].mDataByteSize, audio->inputDeviceBufferList->mBuffers[0].mNumberChannels);
#if 1
        spDebug(80, "inputConverterCallback",
                "inputConvertedOffset = %d, inputDeviceBufferList->mNumberBuffers = %d\n", audio->inputConvertedOffset, audio->inputDeviceBufferList->mNumberBuffers);
#endif
        spDebug(80, "inputConverterCallback",
                "ioData->mNumberBuffers = %d, ioData->mBuffers[0].mNumberChannels = %d\n",
                ioData->mNumberBuffers, ioData->mBuffers[0].mNumberChannels);
    
        for (i = 0; i < ioData->mNumberBuffers; i++) {
            spDebug(100, "inputConverterCallback", "i = %d\n", i);
            if (i < audio->inputDeviceBufferList->mNumberBuffers) {
                ioData->mBuffers[i].mData = audio->inputDeviceBufferList->mBuffers[i].mData + audio->inputConvertedOffset;
                ioData->mBuffers[i].mNumberChannels = audio->inputDeviceBufferList->mBuffers[i].mNumberChannels;
            } else {
                memset(ioData->mBuffers[i].mData, 0, bytesCopied);
                ioData->mBuffers[i].mNumberChannels = audio->inputDeviceBufferList->mBuffers[audio->inputDeviceBufferList->mNumberBuffers-1].mNumberChannels;
            }
            ioData->mBuffers[i].mDataByteSize = bytesCopied;
        }

        audio->inputConvertedOffset += bytesCopied;
    }
    
    if (outDataPacketDescription) {
        *outDataPacketDescription = NULL;
    }
    
    spDebug(80, "inputConverterCallback", "done: inputConvertedOffset = %d\n", audio->inputConvertedOffset);
            
    return noErr;
}

static  OSStatus inputCallback(void *inRefCon,
                               AudioUnitRenderActionFlags *ioActionFlags,
                               const AudioTimeStamp *inTimeStamp,
                               UInt32 inBusNumber,
                               UInt32 inNumberFrames,
                               AudioBufferList *ioData)
{
    OSStatus err = noErr;
    AudioBufferList *bufferList;
    UInt32 c;
    UInt32 numberFrames, convertedFrames, bufferFrames, outputBufferSize;
    UInt32 numberFramesByBufferSize;
    UInt32 propertySize;
    UInt32 bytesCopied, bytesCopied2, bytesCopied3;
    spBool setEventFlag = SP_FALSE;
    spBool stopUnit = SP_FALSE;
    spAudio audio = (spAudio)inRefCon;

    spDebug(80, "inputCallback", "inNumberFrames = %d, mBytesPerFrame = %d, mSampleTime = %f\n",
            inNumberFrames, audio->systemFormat.mBytesPerFrame, inTimeStamp->mSampleTime);

    spLockMutex(audio->inputMutex);
    spDebug(80, "inputCallback", "after spLockMutex\n");
    
    if (audio->inputUnitInitialized == SP_FALSE) {
        spDebug(10, "inputCallback", "inputUnitInitialized == SP_FALSE\n");
        spUnlockMutex(audio->inputMutex);
        return noErr;
    }
    
    audio->inputSampleTime = inTimeStamp->mSampleTime;
    if (audio->inputInitialSampleTime < 0.0) {
        audio->inputInitialSampleTime = inTimeStamp->mSampleTime;
    }

    updateAudioBufferList(audio->inputDeviceBufferList, &audio->inputDeviceBufferAllocSize,
                          inNumberFrames * audio->systemFormat.mBytesPerFrame, SP_TRUE);
    
    if ((err = AudioUnitRender(audio->inputUnit, ioActionFlags, inTimeStamp, inBusNumber, inNumberFrames,
                               audio->inputDeviceBufferList)) != noErr) {
        spDebug(10, "inputCallback", "AudioUnitRender: err = %d\n", err);
        spUnlockMutex(audio->inputMutex);
        return err;
    }
    spDebug(80, "inputCallback", "AudioUnitRender done: inNumberFrames = %d, inputDeviceBufferList size = %d\n",
            inNumberFrames, audio->inputDeviceBufferList->mBuffers[0].mDataByteSize);

    propertySize = sizeof(outputBufferSize);
    outputBufferSize = inNumberFrames * audio->systemFormat.mBytesPerFrame;
    if ((err = AudioConverterGetProperty(audio->inputConverter,
                                         kAudioConverterPropertyCalculateOutputBufferSize,
                                         &propertySize,
                                         &outputBufferSize)) != noErr) {
        spDebug(10, "inputCallback", "AudioConverterGetProperty of kAudioConverterPropertyCalculateOutputBufferSize: err = %d\n", err);
    }
    spDebug(80, "inputCallback", "outputBufferSize = %d, bufferSizeForInput = %d, inputDeviceBufferSize = %d\n",
            outputBufferSize, audio->bufferSizeForInput, audio->inputDeviceBufferSize);

#if 0
    {
        UInt32 inputDeviceBufferSize = outputBufferSize;
        UInt32 inputDeviceFrames;
        
        if ((err = AudioConverterGetProperty(audio->inputConverter,
                                             kAudioConverterPropertyCalculateInputBufferSize,
                                             &propertySize,
                                             &inputDeviceBufferSize)) != noErr) {
            spDebug(10, "inputCallback", "AudioConverterGetProperty of kAudioConverterPropertyCalculateInputBufferSize: err = %d\n", err);
        }
        inputDeviceFrames = inputDeviceBufferSize / audio->systemFormat.mBytesPerFrame;
        spDebug(80, "inputCallback", "inputDeviceBufferSize = %d, inputDeviceFrames = %d, inNumberFrames = %d\n",
                inputDeviceBufferSize, inputDeviceFrames, inNumberFrames);
    }
#endif

    audio->inputConvertedOffset = 0;

    memset(audio->inputConversionBuffer, 0, 2 * audio->bufferSizeForInput * audio->currentFormat.mChannelsPerFrame);

    bufferList = audio->inputConversionBufferList;
    for (c = 0; c < bufferList->mNumberBuffers; c++) {
        bufferList->mBuffers[c].mNumberChannels = 1;
        bufferList->mBuffers[c].mData = audio->inputConversionBuffer + c * 2 * audio->bufferSizeForInput;
    }

    bufferFrames = outputBufferSize / audio->currentFormat.mBytesPerFrame;
    numberFramesByBufferSize = 2 * audio->bufferSizeForInput / audio->currentFormat.mBytesPerFrame;
    convertedFrames = 0;
    spDebug(80, "inputCallback", "bufferFrames = %d, inNumberFrames = %d, numberFramesByBufferSize = %d, currentFormat.mBytesPerFrame = %d\n",
            bufferFrames, inNumberFrames, numberFramesByBufferSize, audio->currentFormat.mBytesPerFrame);
    
    do {
        numberFrames = MIN(bufferFrames - convertedFrames, numberFramesByBufferSize);
        for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
            bufferList->mBuffers[c].mDataByteSize = numberFrames * audio->currentFormat.mBytesPerFrame;
        }
        spDebug(80, "inputCallback", "numberFrames = %d / %d, mDataByteSize = %d\n", 
                numberFrames, bufferFrames, bufferList->mBuffers[0].mDataByteSize);
        
        err = AudioConverterFillComplexBuffer(audio->inputConverter,
                                              inputConverterCallback,
                                              inRefCon,
                                              &numberFrames,
                                              bufferList,
                                              0);

        if (err != noErr) {
            spDebug(10, "inputCallback", "AudioConverterFillComplexBuffer: err = %d\n", err);
            spUnlockMutex(audio->inputMutex);
            return err;
        }
        spDebug(80, "inputCallback", "after AudioConverterFillComplexBuffer: inNumberFrames = %d, numberFrames = %d\n",
                inNumberFrames, numberFrames);

        bytesCopied = numberFrames * audio->currentFormat.mBytesPerFrame;
        spDebug(80, "inputCallback",
                "inNumberFrames = %d, convertedFrames = %d, numberFrames = %d, bytesCopied = %d\n", inNumberFrames, convertedFrames, numberFrames, bytesCopied);
        spDebug(80, "inputCallback",
                "inputBufferSize = %d, inputBufferOffset = %d\n", audio->inputBufferSize, audio->inputBufferOffset);

        bytesCopied2 = MIN(bytesCopied, audio->inputBufferSize - audio->inputBufferOffset);
        for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
            memcpy(audio->inputBuffer + c * audio->inputBufferSize + audio->inputBufferOffset,
                   audio->inputConversionBuffer + c * 2 * audio->bufferSizeForInput, bytesCopied2);
        }
        bytesCopied3 = bytesCopied  - bytesCopied2;
        spDebug(80, "inputCallback", "bytesCopied = %d, bytesCopied2 = %d, bytesCopied3 = %d\n",
                bytesCopied, bytesCopied2, bytesCopied3);
        if (bytesCopied3 > 0) {
            for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
                memcpy(audio->inputBuffer + c * audio->inputBufferSize,
                       audio->inputConversionBuffer + c * 2 * audio->bufferSizeForInput + bytesCopied2, bytesCopied3);
            }
        }
        audio->inputBufferFilled += bytesCopied;
        audio->inputBufferOffset += bytesCopied;
        if (audio->inputBufferOffset >= audio->inputBufferSize) {
            audio->inputBufferOffset -= audio->inputBufferSize;
            spDebug(80, "inputCallback", "reset inputBufferOffset = %d\n", audio->inputBufferOffset);
        }
    
        convertedFrames += numberFrames;
        spDebug(80, "inputCallback", "convertedFrames = %d / %d\n", convertedFrames, bufferFrames);
    } while (convertedFrames < bufferFrames);

    spDebug(80, "inputCallback", "loop finished: convertedFrames = %d / %d, inputBufferFilled = %d, audio->bufferSizeForInput = %d\n",
            convertedFrames, bufferFrames, audio->inputBufferFilled, audio->bufferSizeForInput);
    
    if (audio->inputBufferFilled >= audio->bufferSizeForInput) {
        if (audio->inputBufferFilled >= audio->inputBufferSize) {
            spDebug(80, "inputCallback", "**** stop inputUnit ****, inputCallbackCount = %d\n", audio->inputCallbackCount);
            if (audio->inputCallbackCount > 0 && audio->doStop == SP_FALSE && audio->inputUnitStarted == SP_TRUE) {
                stopUnit = SP_TRUE;
            }
        }
        setEventFlag = SP_TRUE;
    }
        
    audio->inputCallbackCount++;
    
    spUnlockMutex(audio->inputMutex);

    if (stopUnit == SP_TRUE) {
        spDebug(80, "inputCallback", "call AudioOutputUnitStop\n");
        audio->inputUnitStarted = SP_FALSE;
        AudioOutputUnitStop(audio->inputUnit);
        spDebug(80, "inputCallback", "AudioOutputUnitStop done\n");
    }

    if (audio->doStop == SP_TRUE || setEventFlag == SP_TRUE) {
        spDebug(80, "inputCallback", "call spSetEvent\n");
        spSetEvent(audio->inputEvent);
    }
    
    spDebug(80, "inputCallback", "done\n");
    
    return noErr;
}

static OSStatus outputConverterCallback(AudioConverterRef inAudioConverter, 
                                        UInt32 *ioNumberDataPackets, 
                                        AudioBufferList *ioData, 
                                        AudioStreamPacketDescription **outDataPacketDescription, 
                                        void *inUserData)
{
    UInt32 c;
    UInt32 bytesCopied = 0;
    spAudio audio = (spAudio)inUserData;

    spDebug(80, "outputConverterCallback", "ioData = %ld\n", (long)ioData);
    
    ioData->mNumberBuffers = audio->currentFormat.mChannelsPerFrame;
    for (c = 0; c < ioData->mNumberBuffers; c++) {
        spDebug(80, "outputConverterCallback", "ioData->mBuffers[%d].mData = %ld\n", c, (long)ioData->mBuffers[c].mData);
        ioData->mBuffers[c].mData = NULL;
        ioData->mBuffers[c].mDataByteSize = 0;
        ioData->mBuffers[c].mNumberChannels = 1;
    }
    
    spDebug(80, "outputConverterCallback", "*ioNumberDataPackets = %d, audio->outputBufferFilled = %d\n",
            *ioNumberDataPackets, audio->outputBufferFilled);
    
    if (*ioNumberDataPackets > 0) {
        bytesCopied = *ioNumberDataPackets * audio->currentFormat.mBytesPerPacket;
        spDebug(80, "outputConverterCallback", "original bytesCopied = %d / %d\n",
                bytesCopied, audio->bufferSizeForOutput);

        memset(audio->outputConversionBuffer, 0, bytesCopied * audio->currentFormat.mChannelsPerFrame);
        
        spDebug(80, "outputConverterCallback", "audio->outputBufferFilled = %d, outputBufferOffset = %d / %d\n",
                audio->outputBufferFilled, audio->outputBufferOffset, audio->outputBufferSize);

        spDebug(80, "outputConverterCallback", "outputBufferWritePos = %d / %d\n",
                audio->outputBufferWritePos, audio->outputBufferSize);
        
        if (audio->outputBufferOffset + bytesCopied > audio->outputBufferSize) {
            UInt32 bytesCopied2, bytesCopied3;

            bytesCopied2 = audio->outputBufferSize - audio->outputBufferOffset;
            if (bytesCopied2 > 0) {
                for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
                    memcpy(audio->outputConversionBuffer + c * 2 * audio->bufferSizeForOutput,
                           audio->outputBuffer + c * audio->outputBufferSize + audio->outputBufferOffset, bytesCopied2);
                }
            }
            bytesCopied3 = bytesCopied - bytesCopied2;
            for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
                memcpy(audio->outputConversionBuffer + c * 2 * audio->bufferSizeForOutput + bytesCopied2,
                       audio->outputBuffer + c * audio->outputBufferSize, bytesCopied3);
            }
            bytesCopied = bytesCopied2 + bytesCopied3;
            spDebug(80, "outputConverterCallback", "bytesCopied2 = %d, bytesCopied3 = %d\n",
                    bytesCopied2, bytesCopied3);
        } else {
            for (c = 0; c < audio->currentFormat.mChannelsPerFrame; c++) {
                memcpy(audio->outputConversionBuffer + c * 2 * audio->bufferSizeForOutput,
                       audio->outputBuffer + c * audio->outputBufferSize + audio->outputBufferOffset, bytesCopied);
            }
        }
        
        for (c = 0; c < ioData->mNumberBuffers; c++) {
            ioData->mBuffers[c].mData = audio->outputConversionBuffer + c * 2 * audio->bufferSizeForOutput;
            ioData->mBuffers[c].mDataByteSize = bytesCopied;
            ioData->mBuffers[c].mNumberChannels = 1;
        }
        
        audio->outputBufferOffset += bytesCopied;
        if (audio->outputBufferOffset >= audio->outputBufferSize) {
            audio->outputBufferOffset -= audio->outputBufferSize;
        }
        audio->outputBufferFilled -= MIN(bytesCopied, audio->outputBufferFilled);
        spDebug(80, "outputConverterCallback", "bytesCopied = %d, outputBufferFilled = %d\n",
                bytesCopied, audio->outputBufferFilled);
    }
    
    if (outDataPacketDescription) {
        *outDataPacketDescription = NULL;
    }
    
    return noErr;
}

static  OSStatus renderCallback(void *inRefCon,
                                AudioUnitRenderActionFlags *ioActionFlags,
                                const AudioTimeStamp *inTimeStamp,
                                UInt32 inBusNumber,
                                UInt32 inNumberFrames,
                                AudioBufferList *ioData)
{
    UInt32 i;
    UInt32 numberFrames, copyFrames, remainFrames;
    UInt32 offset, copySize, ioDataOffset;
    OSStatus err = noErr;
    spAudio audio = (spAudio)inRefCon;
    spBool setEventFlag = SP_FALSE;
    spBool stopUnit = SP_FALSE;

    spDebug(80, "renderCallback", "in\n");
    
    spLockMutex(audio->outputMutex);

    if (audio->outputUnitInitialized == SP_FALSE) {
        spDebug(10, "renderCallback", "outputUnitInitialized == SP_FALSE\n");
        spUnlockMutex(audio->outputMutex);
        return noErr;
    }

    spDebug(80, "renderCallback",
            "inNumberFrames = %d, mNumberBuffers = %d, mDataByteSize = %d, outputBufferFilled = %d, mSampleTime = %f\n",
            inNumberFrames, ioData->mNumberBuffers, ioData->mBuffers[0].mDataByteSize,
            audio->outputBufferFilled, inTimeStamp->mSampleTime);

    audio->outputSampleTime = inTimeStamp->mSampleTime;
    if (audio->outputInitialSampleTime < 0.0) {
        audio->outputInitialSampleTime = inTimeStamp->mSampleTime;
    }
    
    remainFrames = inNumberFrames;
    ioDataOffset = 0;
    
    do {
        spDebug(100, "renderCallback", "remainFrames = %d, outputConvertedFrames = %d\n",
                remainFrames, audio->outputConvertedFrames);
        
        if (audio->outputConvertedFrames > 0) {
            copyFrames = MIN(remainFrames, audio->outputConvertedFrames - audio->outputConvertedOffset);
            copySize = copyFrames * audio->systemFormat.mBytesPerFrame;
            offset = audio->outputConvertedOffset * audio->systemFormat.mBytesPerFrame;
            spDebug(80, "renderCallback", "copyFrames = %d, copySize = %d, offset = %d, mNumberBuffers = %d (%d), mDataByteSize = %d\n",
                    copyFrames, copySize, offset, audio->outputDeviceBufferList->mNumberBuffers, ioData->mNumberBuffers,
                    audio->outputDeviceBufferList->mBuffers[0].mDataByteSize);

            for (i = 0; i < ioData->mNumberBuffers; i++) {
                if (i < audio->outputDeviceBufferList->mNumberBuffers) {
                    spDebug(80, "renderCallback", "i = %d, mDataByteSize = %d, mNumberChannels = %d\n",
                            i, ioData->mBuffers[i].mDataByteSize, ioData->mBuffers[i].mNumberChannels);
                    memcpy(ioData->mBuffers[i].mData + ioDataOffset, audio->outputDeviceBufferList->mBuffers[i].mData + offset, copySize);
                    ioData->mBuffers[i].mDataByteSize = ioDataOffset + copySize;
                    ioData->mBuffers[i].mNumberChannels = audio->outputDeviceBufferList->mBuffers[i].mNumberChannels;
                } else {
                    memset(ioData->mBuffers[i].mData + ioDataOffset, 0, copySize);
                    ioData->mBuffers[i].mDataByteSize = ioDataOffset + copySize;
                    ioData->mBuffers[i].mNumberChannels = audio->outputDeviceBufferList->mBuffers[audio->outputDeviceBufferList->mNumberBuffers-1].mNumberChannels;
                }
            }
            ioDataOffset += copySize;
            remainFrames -= copyFrames;
            audio->outputConvertedOffset += copyFrames;
            spDebug(80, "renderCallback", "outputConvertedOffset = %d, outputConvertedFrames = %d\n",
                    audio->outputConvertedOffset, audio->outputConvertedFrames);
            
            if (audio->outputConvertedOffset >= audio->outputConvertedFrames) {
                audio->outputConvertedOffset = 0;
                audio->outputConvertedFrames = 0;
            }
        }
        
        if (remainFrames == 0) {
            break;
        }
        
        if (audio->outputConvertedFrames == 0) {
            if (audio->outputBufferFilled > 0) {
                numberFrames = audio->bufferSizeForOutput / audio->currentFormat.mBytesPerFrame;
                numberFrames = MIN(numberFrames, inNumberFrames);
                spDebug(80, "renderCallback",
                        "numberFrames before conversion = %d, inNumberFrames = %d, outputDeviceBufferList->mBuffers[0].mDataByteSize = %d\n",
                        numberFrames, inNumberFrames, audio->outputDeviceBufferList->mBuffers[0].mDataByteSize);

                updateAudioBufferList(audio->outputDeviceBufferList, &audio->outputDeviceBufferAllocSize,
                                      numberFrames * audio->systemFormat.mBytesPerFrame, SP_TRUE);
            
                audio->outputConvertedFrames = 0;
                audio->outputConvertedOffset = 0;
        
                err = AudioConverterFillComplexBuffer(audio->outputConverter,
                                                      outputConverterCallback,
                                                      inRefCon,
                                                      &numberFrames,
                                                      audio->outputDeviceBufferList,
                                                      0);
        
                if (err != noErr) {
                    spDebug(10, "renderCallback", "AudioConverterFillComplexBuffer: err = %d, numberFrames = %d\n",
                            err, numberFrames);
                    break;
                }
                spDebug(80, "renderCallback",
                        "AudioConverterFillComplexBuffer: numberFrames = %d, outputDeviceBufferList->mBuffers[0].mDataByteSize = %d\n",
                        numberFrames, audio->outputDeviceBufferList->mBuffers[0].mDataByteSize);
                audio->outputConvertedFrames = numberFrames;
            } else {
                spDebug(10, "renderCallback", "**** audio->outputBufferFilled <= 0 ****\n");
#if 1
                break;
#else
                clearAudioBufferList(audio->outputDeviceBufferList);
                audio->outputConvertedFrames = MIN(inNumberFrames, audio->bufferSizeForOutput / audio->currentFormat.mBytesPerFrame);
#endif
            }
        }
    } while (audio->doStop == SP_FALSE);

    if (remainFrames != 0) {
        spDebug(10, "renderCallback", "******* remainFrames = %d / %d, outputBufferFilled = %d ********\n",
                remainFrames, inNumberFrames, audio->outputBufferFilled);
    }
    
    if (remainFrames == inNumberFrames && audio->outputBufferFilled == 0) {
        spDebug(10, "renderCallback", "**** stop outputUnit ****\n");
        if (audio->outputCallbackCount > 0 && audio->doStop == SP_FALSE && audio->outputUnitStarted == SP_TRUE)  {
            stopUnit = SP_TRUE;
        }
    }
        
    audio->outputCallbackCount++;

    if (audio->outputWaiting == SP_TRUE 
        || (audio->outputBufferFilled <= audio->outputBufferSize - audio->bufferSizeForOutput)) {
        setEventFlag = SP_TRUE;
    }

    spUnlockMutex(audio->outputMutex);
    
    if (stopUnit == SP_TRUE) {
        spDebug(80, "renderCallback", "call AudioOutputUnitStop\n");
        audio->outputUnitStarted = SP_FALSE;
        AudioOutputUnitStop(audio->outputUnit);
        spDebug(80, "renderCallback", "AudioOutputUnitStop done\n");
    }
    
    if (audio->doStop == SP_TRUE || setEventFlag == SP_TRUE) {
        spDebug(80, "renderCallback", "call spSetEvent(audio->outputEvent)\n");
        spSetEvent(audio->outputEvent);
        spDebug(80, "renderCallback", "call spSetEvent(audio->outputEvent) done\n");
    }
    
    spDebug(80, "renderCallback", "done\n");
    
    return noErr;
}

spBool spInitAudioDriverArch(spAudio audio, const char *driver_name, void *data)
{
    spDebug(80, "spInitAudioDriverArch", "in\n");
    
    audio->inputUnitInitialized = SP_FALSE;
    audio->inputUnitStarted = SP_FALSE;
    audio->outputUnitInitialized = SP_FALSE;
    audio->outputUnitStarted = SP_FALSE;
    audio->inputEnabled = SP_FALSE;
    audio->outputEnabled = SP_FALSE;

    audio->inputEvent = spCreateEvent(SP_TRUE, SP_FALSE);
    audio->outputEvent = spCreateEvent(SP_TRUE, SP_TRUE);
    audio->inputMutex = spCreateMutex(NULL);
    audio->outputMutex = spCreateMutex(NULL);
    
    audio->inputUnit = NULL;
    audio->outputUnit = NULL;
#if !defined(IPHONE)
    audio->inputDevice = 0;
    audio->outputDevice = 0;
#endif
    audio->component = NULL;
    
    memset(&audio->systemFormat, 0, sizeof(audio->systemFormat));
    memset(&audio->currentFormat, 0, sizeof(audio->currentFormat));
    
    audio->inputConverter = NULL;
    audio->outputConverter = NULL;

    audio->inputDeviceBufferList = NULL;
    audio->inputConversionBufferList = NULL;
    audio->outputDeviceBufferList = NULL;
    audio->inputBuffer = NULL;
    audio->outputBuffer = NULL;
    audio->inputDeviceBufferAllocSize = audio->outputDeviceBufferAllocSize = 0;
    audio->inputDeviceBufferSize = audio->outputDeviceBufferSize = 0;
    audio->inputBufferSize = audio->outputBufferSize = 0;
    audio->bufferSizeForInput = audio->bufferSizeForOutput = 0;
    
    audio->inputConversionBuffer = NULL;
    audio->outputConversionBuffer = NULL;
    
    resetBufferCounter(audio);
    
    memset(&audio->cd, 0, sizeof(audio->cd));
    audio->cd.componentType = kAudioUnitType_Output;
    audio->cd.componentManufacturer = kAudioUnitManufacturer_Apple;
    
#if defined(IPHONE)
    audio->cd.componentSubType = kAudioUnitSubType_RemoteIO;
    audio->component = AudioComponentFindNext(NULL, &audio->cd);
    if (audio->component != NULL) {
        AudioComponentInstanceNew(audio->component, &audio->inputUnit);
        AudioComponentInstanceNew(audio->component, &audio->outputUnit);
    }
#else
    audio->cd.componentSubType = /*kAudioUnitSubType_DefaultOutput*/kAudioUnitSubType_HALOutput;
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1060
    audio->component = AudioComponentFindNext(NULL, &audio->cd);
    if (audio->component != NULL) {
        AudioComponentInstanceNew(audio->component, &audio->inputUnit);
        AudioComponentInstanceNew(audio->component, &audio->outputUnit);
    }
#else
    audio->component = FindNextComponent(NULL, &audio->cd);
    if (audio->component != NULL) {
        OpenAComponent(audio->component, &audio->inputUnit);
        OpenAComponent(audio->component, &audio->outputUnit);
    }
#endif
#endif

    if (audio->inputUnit == NULL || audio->outputUnit == NULL) {
        spDebug(10, "spInitAudioDriverArch", "open compnent failed\n");
        spFreeAudioDriverArch(audio);
    
        return SP_FALSE;
    } else {
        audio->systemFormat.mSampleRate = 44100.0;
        audio->systemFormat.mFormatID = kAudioFormatLinearPCM;
        audio->systemFormat.mFormatFlags = SP_SYSTEM_AUDIO_UNIT_FORMAT_FLAGS;
        audio->systemFormat.mFramesPerPacket = 1;
        audio->systemFormat.mChannelsPerFrame = 1;
        audio->systemFormat.mBitsPerChannel = 8 * sizeof(spAudioUnitSampleType);
        audio->systemFormat.mBytesPerFrame = sizeof(spAudioUnitSampleType);
        audio->systemFormat.mBytesPerPacket = audio->systemFormat.mBytesPerFrame;

        audio->currentFormat.mSampleRate = 8000.0;
        audio->currentFormat.mFormatID = kAudioFormatLinearPCM;
        audio->currentFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian
            | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
        audio->currentFormat.mFramesPerPacket = 1;
        audio->currentFormat.mChannelsPerFrame = 1;
        audio->currentFormat.mBitsPerChannel = 16;
        audio->currentFormat.mBytesPerPacket = audio->currentFormat.mBytesPerFrame = 2;
        
        spDebug(80, "spInitAudioDriverArch", "done\n");
        
        return SP_TRUE;
    }
}

static spBool openAudio(spAudio audio)
{
    OSStatus err = noErr;
    UInt32 flag;

    spDebug(10, "openAudio", "inputEnabled = %d, outputEnabled = %d\n", audio->inputEnabled, audio->outputEnabled);
    
    if (audio->inputUnit == NULL && audio->outputUnit == NULL) {
        return SP_FALSE;
    }
    
#if defined(IPHONE)
    {
        UInt32 bufferSize;
        Float32 bufferLengthSec;

        if (audio->outputEnabled == SP_TRUE) {
            bufferSize = audio->outputDeviceBufferSize;
        } else {
            bufferSize = audio->inputDeviceBufferSize;
        }
        bufferLengthSec = (Float32)bufferSize / (Float32)audio->samp_rate;
        spDebug(10, "openAudio", "bufferSize = %d, bufferLengthSec = %f\n", bufferSize, bufferLengthSec);
        
        if ((err = AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareIOBufferDuration, 
                                           sizeof(bufferLengthSec), &bufferLengthSec)) != noErr) {
            spDebug(10, "openAudio", "AudioSessionSetProperty of kAudioSessionProperty_PreferredHardwareIOBufferDuration failed: err = %d\n", err);
        }
    }
#endif

    if (audio->inputEnabled == SP_FALSE) {
        flag = 0;
        AudioUnitSetProperty(audio->inputUnit, 
                             kAudioOutputUnitProperty_EnableIO, 
                             kAudioUnitScope_Input, 
                             kInputBus,
                             &flag, 
                             sizeof(flag));
    }
        
    if (audio->outputEnabled == SP_FALSE) {
        flag = 0;
        AudioUnitSetProperty(audio->outputUnit, 
                             kAudioOutputUnitProperty_EnableIO, 
                             kAudioUnitScope_Output, 
                             kOutputBus,
                             &flag, 
                             sizeof(flag));
    }

    audio->inputUnitInitialized = audio->outputUnitInitialized = SP_FALSE;
    
    if (audio->inputEnabled == SP_TRUE) {
        if ((err = AudioUnitInitialize(audio->inputUnit)) == noErr) {
            audio->inputUnitInitialized = SP_TRUE;

            audio->inputDeviceBufferAllocSize = audio->inputDeviceBufferSize + SP_AUDIO_AU_BUFFER_MARGIN;
            audio->inputDeviceBufferList = allocAudioBufferList(audio->systemFormat.mChannelsPerFrame,
                                                                audio->inputDeviceBufferAllocSize);
            audio->inputConversionBufferList = allocAudioBufferList(audio->currentFormat.mChannelsPerFrame, 0);
        } else {
            spDebug(10, "openAudio", "AudioUnitInitialize of input: err = %d\n", err);
            return SP_FALSE;
        }
    }

    if (audio->outputEnabled == SP_TRUE) {
        if ((err = AudioUnitInitialize(audio->outputUnit)) == noErr) {
            audio->outputUnitInitialized = SP_TRUE;

            audio->outputDeviceBufferAllocSize = audio->outputDeviceBufferSize + SP_AUDIO_AU_BUFFER_MARGIN;
            audio->outputDeviceBufferList = allocAudioBufferList(audio->systemFormat.mChannelsPerFrame,
                                                                 audio->outputDeviceBufferAllocSize);
        } else {
            spDebug(10, "openAudio", "AudioUnitInitialize of output: err = %d\n", err);
            return SP_FALSE;
        }
    }
            
    allocInternalBuffer(audio);
    allocConversionBuffer(audio);
            
    resetBufferCounter(audio);
            
    spDebug(10, "openAudio", "done\n");
            
    return SP_TRUE;
}

static spBool startInputAudio(spAudio audio)
{
    OSStatus err = noErr;
    
    spDebug(80, "startInputAudio", "in\n");
    
    spLockMutex(audio->inputMutex);
    audio->inputCallbackCount = 0;
    spUnlockMutex(audio->inputMutex);
    
    spDebug(80, "startInputAudio", "call AudioOutputUnitStart\n");
    err = AudioOutputUnitStart(audio->inputUnit);
    audio->inputUnitStarted = SP_TRUE;
    spDebug(80, "startInputAudio", "AudioOutputUnitStart done\n");
    
    if (err != noErr) {
        spDebug(10, "startInputAudio", "AudioOutputUnitStart: err = %d\n", err);
        audio->inputUnitStarted = SP_FALSE;
        return SP_FALSE;
    }
    
    spDebug(80, "startInputAudio", "done\n");
    
    return SP_TRUE;
}

static spBool startOutputAudio(spAudio audio)
{
    OSStatus err = noErr;
    
    spDebug(80, "startOutputAudio", "in\n");
    
    spLockMutex(audio->outputMutex);
    audio->outputCallbackCount = 0;
    spUnlockMutex(audio->outputMutex);
    
    spDebug(80, "startOutputAudio", "call AudioOutputUnitStart\n");
    err = AudioOutputUnitStart(audio->outputUnit);
    audio->outputUnitStarted = SP_TRUE;
    spDebug(80, "startOutputAudio", "AudioOutputUnitStart done\n");
    
    if (err != noErr) {
        spDebug(10, "startOutputAudio", "AudioOutputUnitStart: err = %d\n", err);
        audio->outputUnitStarted = SP_FALSE;
        return SP_FALSE;
    }
    
    spDebug(80, "startOutputAudio", "done\n");
    
    return SP_TRUE;
}

spBool spGetNumAudioDeviceArch(spAudio audio, int *num_device)
{
    *num_device = 1;
    return SP_TRUE;
}

const char *spGetAudioDeviceNameArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return "Default";
    }
    return NULL;
}

spBool spSelectAudioDeviceArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return SP_TRUE;
    }
    return SP_FALSE;
}

spBool spSetAudioSampleBitArch(spAudio audio)
{
    if (audio->samp_bit == 33 || audio->samp_bit == 64) {
        audio->currentFormat.mFormatFlags = kAudioFormatFlagIsFloat | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
        return SP_TRUE;
    } else if (audio->samp_bit == 8 || audio->samp_bit == 16
               || audio->samp_bit == 24 || audio->samp_bit == 32) {
        audio->currentFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spSetAudioSampleRateArch(spAudio audio)
{
    audio->currentFormat.mSampleRate = audio->samp_rate;
    
    return SP_TRUE;
}

spBool spSetAudioChannelArch(spAudio audio)
{
    audio->systemFormat.mChannelsPerFrame = audio->currentFormat.mChannelsPerFrame = audio->num_channel;
    /*audio->currentFormat.mChannelsPerFrame = audio->num_channel;*/
    
    return SP_TRUE;
}

spBool spSetAudioBufferSizeArch(spAudio audio)
{
    /*audio->buffer_size = MAX(audio->buffer_size, SP_AUDIO_AU_MIN_BUFFER_SIZE);*/
    return SP_TRUE;
}

spBool spSetAudioNumBufferArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioBlockingModeArch(spAudio audio)
{
    return SP_TRUE;
}

spAudioCallbackType spSetAudioCallbackFuncArch(spAudio audio)
{
    return (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK);
}

#if defined(IPHONE)
OSStatus initAudioSession(UInt32 category, spBool read_flag)
{
    OSStatus err = noErr;

    spDebug(80, "initAudioSession", "category = %d\n", category);
    
    if ((err = AudioSessionInitialize(NULL, NULL, NULL, NULL)) == noErr) {
        if (read_flag) {
            UInt32 inputIsAvailable = 0;
            UInt32 propertySize;

            propertySize = sizeof(inputIsAvailable);
            if ((err = AudioSessionGetProperty(kAudioSessionProperty_AudioInputAvailable, &propertySize,&inputIsAvailable)) != noErr
                || inputIsAvailable == 0) {
                spDebug(80, "initAudioSession", "AudioSessionGetProperty kAudioSessionProperty_AudioInputAvailable failed: err = %d\n", err);
                return kAudioUnitErr_NoConnection;
            }
        }
        
        if ((err = AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(category), &category)) == noErr) {
            if ((err = AudioSessionSetActive(true)) == noErr) {
                spDebug(80, "initAudioSession", "AudioSessionSetActive successfully done\n");
                return noErr;
            } else {
                spDebug(10, "initAudioSession", "AudioSessionSetActive failed: err = %d\n", err);
            }
        } else {
            spDebug(10, "initAudioSession", "AudioSessionSetProperty failed: err = %d\n", err);
        }
    } else {
        spDebug(10, "initAudioSession", "AudioSessionInitialize failed: err = %d\n", err);
    }

    return err;
}

OSStatus setPreferredHardwareSampleRate(Float64 *sampleRate)
{
    OSStatus err = noErr;
    UInt32 propertySize;

    spDebug(80, "setPreferredHardwareSampleRate", "original: sampleRate = %f\n", *sampleRate);
    
    if ((err = AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareSampleRate, sizeof(Float64), sampleRate)) != noErr) {
        spDebug(10, "setPreferredHardwareSampleRate", "AudioSessionSetProperty kAudioSessionProperty_PreferredHardwareSampleRate failed: err = %d\n", err);
    }

    propertySize = sizeof(Float64);
    AudioSessionGetProperty(kAudioSessionProperty_PreferredHardwareSampleRate, &propertySize, sampleRate);
    spDebug(80, "setPreferredHardwareSampleRate", "AudioSessionGetProperty kAudioSessionProperty_PreferredHardwareSampleRate: sampleRate = %f\n", *sampleRate);
    
    return err;
}
#else
AudioObjectID getDefaultDevice(spBool outputFlag)
{
    OSStatus err = noErr;
    AudioObjectID device;
    UInt32 propertySize;
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
    AudioObjectPropertyAddress address = { kAudioHardwarePropertyDefaultInputDevice,
                                           kAudioObjectPropertyScopeGlobal,
                                           kAudioObjectPropertyElementMaster };
#endif

    if (outputFlag) {
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
        address.mSelector = kAudioHardwarePropertyDefaultOutputDevice;
#endif
    }
    
    propertySize = sizeof(AudioObjectID);

#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
    err = AudioObjectGetPropertyData(kAudioObjectSystemObject, &address, 0, NULL, &propertySize, &device);
#else
    err = AudioHardwareGetProperty(outputFlag ? kAudioHardwarePropertyDefaultOutputDevice : kAudioHardwarePropertyDefaultInputDevice,
                                   &propertySize, &device);
#endif

    if (err == noErr) {
        return device;
    } else {
        spDebug(10, "getDefaultDevice", "AudioHardwareGetProperty: err = %d\n", err);
    }
    
    return 0;
}

spBool setupDefaultDevice(spAudio audio, AudioUnit unit, AudioObjectID device, UInt32 bufferSize, spBool outputFlag)
{
    OSStatus err = noErr;
    AudioUnitElement bus = kInputBus;
    UInt32 bufferFrameSize;
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
    AudioObjectPropertyAddress address = { kAudioDevicePropertyBufferFrameSize,
                                           kAudioObjectPropertyScopeGlobal,
                                           kAudioObjectPropertyElementMaster };
#endif

    if (outputFlag) {
        bus = kOutputBus;
    }
    
    spDebug(10, "setupDefaultDevice", "device = %ld\n", (long)device);
        
    if ((err = AudioUnitSetProperty(unit, kAudioOutputUnitProperty_CurrentDevice,
                                    kAudioUnitScope_Global, bus,
                                    &device, sizeof(AudioObjectID))) == noErr) {
        bufferFrameSize = bufferSize / audio->systemFormat.mBytesPerFrame;

#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
        err = AudioObjectSetPropertyData(device, &address, 0, NULL, sizeof(bufferFrameSize), &bufferFrameSize);
#else
        err = AudioDeviceSetProperty(device, NULL, 0, false, kAudioDevicePropertyBufferFrameSize,
                                     sizeof(bufferFrameSize), &bufferFrameSize);
#endif
        if (err != noErr) {
            spDebug(10, "setupDefaultDevice",
                    "AudioDevicesSetProperty of kAudioDevicePropertyBufferFrameSize: err = %d\n", err);
        }

        return SP_TRUE;
    } else {
        spDebug(10, "setupDefaultDevice", "AudioUnitSetProperty of kAudioOutputUnitProperty_CurrentDevice: err = %d\n", err);
    }
    
    return SP_FALSE;
}
#endif

spBool spOpenInputAudioDeviceArch(spAudio audio)
{
    int i;
    OSStatus err = noErr;
    UInt32 flag = 1;
    UInt32 propertySize;
    UInt32 bufferFrameSize;
    AURenderCallbackStruct rcs;
    Float64 defaultSampleRate;
#if defined(IPHONE)
    UInt32 category = kAudioSessionCategory_PlayAndRecord;

#if 0
     /* sometimes doesn't work on iOS 8 (maybe after playing audio finished) */
    if (audio->only_flag) {
        category = kAudioSessionCategory_RecordAudio;
    }
#endif

    if ((err = initAudioSession(category, SP_TRUE)) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "initAudioSession failed: err = %d\n", err);
        return SP_FALSE;
    }
#endif

    if ((err = AudioUnitSetProperty(audio->inputUnit, 
                                    kAudioOutputUnitProperty_EnableIO, 
                                    kAudioUnitScope_Input, 
                                    kInputBus,
                                    &flag, 
                                    sizeof(flag))) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioUnitSetProperty of kAudioOutputUnitProperty_EnableIO: err = %d\n", err);
        return SP_FALSE;
    }

    updateFormatInformation(audio);

    defaultSampleRate = audio->currentFormat.mSampleRate;
    spDebug(10, "spOpenInputAudioDeviceArch", "input defaultSampleRate = %f\n", defaultSampleRate);
    
#if defined(IPHONE)
    setPreferredHardwareSampleRate(&defaultSampleRate);
#else
    if ((audio->inputDevice = getDefaultDevice(SP_FALSE)) == 0) {
        spDebug(10, "spOpenInputAudioDeviceArch", "getDefaultDevice failed\n");
        return SP_FALSE;
    }
    {
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
        AudioObjectPropertyAddress address = { kAudioDevicePropertyNominalSampleRate,
                                               kAudioObjectPropertyScopeGlobal,
                                               kAudioObjectPropertyElementMaster };

        propertySize = sizeof(defaultSampleRate);
        if ((err = AudioObjectGetPropertyData(audio->inputDevice, &address, 0, NULL, &propertySize, &defaultSampleRate)) != noErr) {
            spDebug(10, "spOpenInputAudioDeviceArch", "AudioObjectGetPropertyData of kAudioDevicePropertyNominalSampleRate: err = %d\n", err);
        }
#else
        propertySize = sizeof(defaultSampleRate);
        if ((err = AudioDeviceGetProperty(device, NULL, 0, false, kAudioDevicePropertyNominalSampleRate, 
                                          &propertySize, &defaultSampleRate)) != noErr) {
            spDebug(10, "spOpenInputAudioDeviceArch", "AudioDeviceGetProperty of kAudioDevicePropertyNominalSampleRate: err = %d\n", err);
        }
#endif
        spDebug(10, "spOpenInputAudioDeviceArch", "defaultSampleRate = %f\n", defaultSampleRate);
    }
#endif
    
    audio->systemFormat.mSampleRate = defaultSampleRate;
    for (i = 0;; i++) {
        spDebug(10, "spOpenInputAudioDeviceArch", "i = %d, mSampleRate = %f\n", i, audio->systemFormat.mSampleRate);
        if ((err = AudioUnitSetProperty(audio->inputUnit, 
                                        kAudioUnitProperty_StreamFormat, 
                                        kAudioUnitScope_Output, 
                                        kInputBus, 
                                        &audio->systemFormat, 
                                        sizeof(audio->systemFormat))) != noErr) {
            spDebug(10, "spOpenInputAudioDeviceArch", "AudioUnitSetProperty of kAudioUnitProperty_StreamFormat: err = %d\n", err);
            if (i > 0) {
                return SP_FALSE;
            }
        } else {
            break;
        }
        
        audio->systemFormat.mSampleRate = 44100.0;
    }

    spDebug(10, "spOpenInputAudioDeviceArch", "updated mSampleRate = %f\n", audio->systemFormat.mSampleRate);
    
    if ((err = AudioConverterNew(&audio->systemFormat, &audio->currentFormat, &audio->inputConverter)) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioConverterNew failed: err = %d\n", err);
        return SP_FALSE;
    }
    
    audio->inputDeviceBufferSize = audio->buffer_size;
    propertySize = sizeof(audio->inputDeviceBufferSize);
    
    if ((err = AudioConverterGetProperty(audio->inputConverter,
                                         kAudioConverterPropertyCalculateInputBufferSize,
                                         &propertySize,
                                         &audio->inputDeviceBufferSize)) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioConverterGetProperty of kAudioConverterPropertyCalculateInputBufferSize: err = %d\n", err);
        AudioConverterDispose(audio->inputConverter);
        return SP_FALSE;
    }
    spDebug(10, "spOpenInputAudioDeviceArch",
            "first kAudioConverterPropertyCalculateInputBufferSize: inputDeviceBufferSize = %d, buffer_size = %d\n",
            audio->inputDeviceBufferSize, audio->buffer_size);
    
    audio->inputDeviceBufferSize = MAX(audio->inputDeviceBufferSize, SP_AUDIO_AU_MIN_BUFFER_SIZE);

    /* buffer size recalculation required for sample rate of higher than 44.1k */
    if ((err = AudioConverterGetProperty(audio->inputConverter,
                                         kAudioConverterPropertyCalculateOutputBufferSize,
                                         &propertySize,
                                         &audio->inputDeviceBufferSize)) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioConverterGetProperty of kAudioConverterPropertyCalculateOutputBufferSize: err = %d\n", err);
        AudioConverterDispose(audio->inputConverter);
        return SP_FALSE;
    }

    audio->bufferSizeForInput = audio->inputDeviceBufferSize;
    spDebug(10, "spOpenInputAudioDeviceArch", "updated bufferSizeForInput = %d, buffer_size = %d\n",
            audio->bufferSizeForInput, audio->buffer_size);

    if ((err = AudioConverterGetProperty(audio->inputConverter,
                                         kAudioConverterPropertyCalculateInputBufferSize,
                                         &propertySize,
                                         &audio->inputDeviceBufferSize)) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioConverterGetProperty of kAudioConverterPropertyCalculateInputBufferSize: err = %d\n", err);
        AudioConverterDispose(audio->inputConverter);
        return SP_FALSE;
    }
        
    spDebug(50, "spOpenInputAudioDeviceArch", "final: bufferSizeForInput = %d (%d), inputDeviceBufferSize = %d\n",
            audio->bufferSizeForInput, audio->buffer_size, audio->inputDeviceBufferSize);

    flag = 0;
    AudioUnitSetProperty(audio->inputUnit, 
                         kAudioOutputUnitProperty_EnableIO, 
                         kAudioUnitScope_Output, 
                         kOutputBus,
                         &flag, 
                         sizeof(flag));
            
#if !defined(IPHONE)
    setupDefaultDevice(audio, audio->inputUnit, audio->inputDevice, audio->inputDeviceBufferSize, SP_FALSE);
#endif

    bufferFrameSize = audio->inputDeviceBufferSize / audio->systemFormat.mBytesPerFrame;
    if ((err = AudioUnitSetProperty(audio->inputUnit, kAudioUnitProperty_MaximumFramesPerSlice,
                                    kAudioUnitScope_Global, kInputBus,
                                    &bufferFrameSize, sizeof(bufferFrameSize))) == noErr) {
    } else {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioUnitSetProperty of kAudioUnitProperty_MaximumFramesPerSlice: err = %d\n", err);
    }
    spDebug(10, "spOpenInputAudioDeviceArch", "bufferFrameSize = %d, systemFormat.mBytesPerFrame = %d, bufferSizeForInput = %d\n",
            bufferFrameSize, audio->systemFormat.mBytesPerFrame, audio->bufferSizeForInput);
            
    memset(&rcs, 0, sizeof(rcs));
        
    rcs.inputProc = inputCallback;
    rcs.inputProcRefCon = audio;
    
    if ((err = AudioUnitSetProperty(audio->inputUnit,
                                    kAudioOutputUnitProperty_SetInputCallback,
                                    kAudioUnitScope_Global,
                                    kInputBus,
                                    &rcs, sizeof(rcs))) != noErr) {
        spDebug(10, "spOpenInputAudioDeviceArch", "AudioUnitSetProperty of kAudioOutputUnitProperty_SetInputCallback: err = %d\n", err);
        return SP_FALSE;
    }
    
    audio->inputEnabled = SP_TRUE;
    audio->doStop = SP_FALSE;
    
    spDebug(50, "spOpenInputAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

spBool spOpenOutputAudioDeviceArch(spAudio audio)
{
    int i;
    OSStatus err = noErr;
    UInt32 flag = 1;
    UInt32 propertySize;
    UInt32 bufferFrameSize;
    AURenderCallbackStruct rcs;
    Float64 defaultSampleRate;
#if defined(IPHONE)
    UInt32 category = kAudioSessionCategory_PlayAndRecord;

    if (audio->only_flag) {
        category = kAudioSessionCategory_MediaPlayback;
    }

    if ((err = initAudioSession(category, SP_FALSE)) != noErr) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "initAudioSession failed: err = %d\n", err);
        return SP_FALSE;
    }
#endif
    
    if (AudioUnitSetProperty(audio->outputUnit, 
                             kAudioOutputUnitProperty_EnableIO, 
                             kAudioUnitScope_Output, 
                             kOutputBus,
                             &flag, 
                             sizeof(flag)) != noErr) {
        return SP_FALSE;
    }

    updateFormatInformation(audio);

    defaultSampleRate = audio->currentFormat.mSampleRate;
    spDebug(10, "spOpenOutputAudioDeviceArch", "input defaultSampleRate = %f\n", defaultSampleRate);
    
#if defined(IPHONE)
    setPreferredHardwareSampleRate(&defaultSampleRate);
#else
    if ((audio->outputDevice = getDefaultDevice(SP_TRUE)) == 0) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "getDefaultDevice failed\n");
        return SP_FALSE;
    }
    {
#if MAC_OS_X_VERSION_MAX_ALLOWED >= 1040
        AudioObjectPropertyAddress address = { kAudioDevicePropertyNominalSampleRate,
                                               kAudioObjectPropertyScopeGlobal,
                                               kAudioObjectPropertyElementMaster };

        propertySize = sizeof(defaultSampleRate);
        if ((err = AudioObjectGetPropertyData(audio->outputDevice, &address, 0, NULL, &propertySize, &defaultSampleRate)) != noErr) {
            spDebug(10, "spOpenOutputAudioDeviceArch", "AudioObjectGetPropertyData of kAudioDevicePropertyNominalSampleRate: err = %d\n", err);
        }
#else
        propertySize = sizeof(defaultSampleRate);
        if ((err = AudioDeviceGetProperty(device, NULL, 0, false, kAudioDevicePropertyNominalSampleRate, 
                                          &propertySize, &defaultSampleRate)) != noErr) {
            spDebug(10, "spOpenOutputAudioDeviceArch", "AudioDeviceGetProperty of kAudioDevicePropertyNominalSampleRate: err = %d\n", err);
        }
#endif
        spDebug(10, "spOpenOutputAudioDeviceArch", "defaultSampleRate = %f\n", defaultSampleRate);
    }
#endif
    
    audio->systemFormat.mSampleRate = defaultSampleRate;
    for (i = 0;; i++) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "mSampleRate = %f\n", audio->systemFormat.mSampleRate);
        if ((err = AudioUnitSetProperty(audio->outputUnit, 
                                        kAudioUnitProperty_StreamFormat, 
                                        kAudioUnitScope_Input, 
                                        kOutputBus, 
                                        &audio->systemFormat, 
                                        sizeof(audio->systemFormat))) != noErr) {
            spDebug(10, "spOpenOutputAudioDeviceArch", "AudioUnitSetProperty of kAudioUnitProperty_StreamFormat: err = %d\n", err);
            if (i > 0 || audio->inputEnabled == SP_TRUE) {
                return SP_FALSE;
            }
        } else {
            AudioStreamBasicDescription format;
            UInt32 size;
            
            spDebug(10, "spOpenOutputAudioDeviceArch", "kAudioUnitProperty_StreamFormat OK: i = %d\n", i);

            size = sizeof(format);
            AudioUnitGetProperty(audio->outputUnit, 
                                 kAudioUnitProperty_StreamFormat, 
                                 kAudioUnitScope_Input, 
                                 kOutputBus, 
                                 &format, 
                                 &size);
            spDebug(10, "spOpenOutputAudioDeviceArch", "format.mSampleRate = %f, format.mBitsPerChannel = %d, format.mBytesPerFrame = %d, format.mFormatFlags = %lx\n",
                    format.mSampleRate, format.mBitsPerChannel, format.mBytesPerFrame, format.mFormatFlags);
            break;
        }
        
        audio->systemFormat.mSampleRate = 44100.0;
    }   

    if ((err = AudioConverterNew(&audio->currentFormat, &audio->systemFormat, &audio->outputConverter)) != noErr) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "AudioConverterNew failed: err = %d\n", err);
        return SP_FALSE;
    }
    
    audio->outputDeviceBufferSize = audio->buffer_size;
    propertySize = sizeof(audio->outputDeviceBufferSize);
    
    if (AudioConverterGetProperty(audio->outputConverter,
                                  kAudioConverterPropertyCalculateOutputBufferSize,
                                  &propertySize,
                                  &audio->outputDeviceBufferSize) != noErr) {
        AudioConverterDispose(audio->outputConverter);
        return SP_FALSE;
    }
    
    audio->outputDeviceBufferSize = MAX(audio->outputDeviceBufferSize, SP_AUDIO_AU_MIN_BUFFER_SIZE);
    spDebug(10, "spOpenOutputAudioDeviceArch", "original: audio->outputDeviceBufferSize = %d, buffer_size = %d\n",
            audio->outputDeviceBufferSize, audio->buffer_size);
    
    /* buffer size recalculation required for sample rate of higher than 44.1k */
    if (AudioConverterGetProperty(audio->outputConverter,
                                  kAudioConverterPropertyCalculateInputBufferSize,
                                  &propertySize,
                                  &audio->outputDeviceBufferSize) != noErr) {
        AudioConverterDispose(audio->outputConverter);
        return SP_FALSE;
    }

    spDebug(10, "spOpenOutputAudioDeviceArch", "kAudioConverterPropertyCalculateInputBufferSize: audio->outputDeviceBufferSize = %d\n",
            audio->outputDeviceBufferSize);

    audio->bufferSizeForOutput = audio->outputDeviceBufferSize;
    spDebug(10, "spOpenOutputAudioDeviceArch", "updated bufferSizeForOutput = %d\n", audio->bufferSizeForOutput);
    
    if (AudioConverterGetProperty(audio->outputConverter,
                                  kAudioConverterPropertyCalculateOutputBufferSize,
                                  &propertySize,
                                  &audio->outputDeviceBufferSize) != noErr) {
        AudioConverterDispose(audio->outputConverter);
        return SP_FALSE;
    }
    
    spDebug(10, "spOpenOutputAudioDeviceArch", "final: bufferSizeForOutput = %d (%d), outputDeviceBufferSize = %d\n",
            audio->bufferSizeForOutput, audio->buffer_size, audio->outputDeviceBufferSize);
    
    flag = 0;
    AudioUnitSetProperty(audio->outputUnit, 
                         kAudioOutputUnitProperty_EnableIO, 
                         kAudioUnitScope_Input, 
                         kInputBus,
                         &flag, 
                         sizeof(flag));
            
#if !defined(IPHONE)
    setupDefaultDevice(audio, audio->outputUnit, audio->outputDevice, audio->outputDeviceBufferSize, SP_TRUE);
#endif

    bufferFrameSize = audio->outputDeviceBufferSize / audio->systemFormat.mBytesPerFrame;
    if ((err = AudioUnitSetProperty(audio->outputUnit, kAudioUnitProperty_MaximumFramesPerSlice,
                                    kAudioUnitScope_Global, kOutputBus,
                                    &bufferFrameSize, sizeof(bufferFrameSize))) == noErr) {
    } else {
        spDebug(10, "spOpenOutputAudioDeviceArch", "AudioUnitSetProperty of kAudioUnitProperty_MaximumFramesPerSlice: err = %d\n", err);
    }
    spDebug(10, "spOpenOutputAudioDeviceArch", "bufferFrameSize = %d, bufferSizeForOutput = %d\n", bufferFrameSize, audio->bufferSizeForOutput);
            
    memset(&rcs, 0, sizeof(rcs));
    
    rcs.inputProc = renderCallback;
    rcs.inputProcRefCon = audio;
    
    if ((err = AudioUnitSetProperty(audio->outputUnit,
                                    kAudioUnitProperty_SetRenderCallback,
                                    kAudioUnitScope_Global,
                                    kOutputBus,
                                    &rcs, sizeof(rcs))) != noErr) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "AudioUnitSetProperty of kAudioUnitProperty_SetRenderCallback: err = %d\n", err);
        return SP_FALSE;
    }

    audio->outputEnabled = SP_TRUE;
    audio->doStop = SP_FALSE;
    
    return SP_TRUE;
}

spBool spCloseAudioDeviceArch(spAudio audio)
{
    UInt32 size;
    Boolean running;
    OSStatus err = noErr;
    
    spDebug(100, "spCloseAudioDeviceArch", "in\n");
    
    if (audio->outputBufferFilled > 0 && audio->outputUnitStarted == SP_FALSE
        && audio->doStop == SP_FALSE) {
        startOutputAudio(audio);
        spDebug(100, "spCloseAudioDeviceArch", "startOutputAudio done\n");
    }
    
    if (audio->outputUnitStarted == SP_TRUE) {
        if (audio->outputEnabled == SP_TRUE) {
            waitOutputAudio(audio);
        }
        audio->outputUnitStarted = SP_FALSE;
        
        spDebug(100, "spCloseAudioDeviceArch", "AudioOutputUnitStop of output unit\n");
        AudioOutputUnitStop(audio->outputUnit);

        do {
            spDebug(100, "spCloseAudioDeviceArch", "ckeck output running...\n");
            running = TRUE;
            size = sizeof(running);
            err = AudioUnitGetProperty(audio->outputUnit,  
                                       kAudioOutputUnitProperty_IsRunning,
                                       kAudioUnitScope_Global,
                                       kOutputBus,
                                       &running,
                                       &size);
            spDebug(100, "spCloseAudioDeviceArch", "output running = %d\n");
            
            if (err != noErr || !running) {
                break;
            }
            
            spMSleep(1);
        } while (running);
        
        spSetEvent(audio->outputEvent);
    }
    spDebug(100, "spCloseAudioDeviceArch", "output unit stopped\n");
    
    audio->doStop = SP_TRUE;
    
    if (audio->inputUnitStarted == SP_TRUE) {
        spDebug(100, "spCloseAudioDeviceArch", "AudioOutputUnitStop of input unit\n");
        AudioOutputUnitStop(audio->inputUnit);
        
        do {
            spDebug(100, "spCloseAudioDeviceArch", "ckeck input running...\n");
            running = TRUE;
            size = sizeof(running);
            err = AudioUnitGetProperty(audio->inputUnit,  
                                       kAudioOutputUnitProperty_IsRunning,
                                       kAudioUnitScope_Global,
                                       kInputBus,
                                       &running,
                                       &size);
            spDebug(100, "spCloseAudioDeviceArch", "input running = %d\n");
            
            if (err != noErr || !running) {
                break;
            }
            
            spMSleep(1);
        } while (running);
        
        audio->inputUnitStarted = SP_FALSE;
        
        spSetEvent(audio->inputEvent);
    }
    spDebug(100, "spCloseAudioDeviceArch", "input unit stopped\n");
    
    spLockMutex(audio->inputMutex);
    if (audio->inputUnitInitialized == SP_TRUE) {
        spDebug(100, "spCloseAudioDeviceArch", "free input buffers\n");
        freeInternalBuffer(audio, SP_FALSE);
        freeConversionBuffer(audio, SP_FALSE);
        
        if (audio->inputConversionBufferList != NULL) {
            spDebug(100, "spCloseAudioDeviceArch", "call freeAudioBufferList for inputConversionBufferList\n");
            free(audio->inputConversionBufferList);
            audio->inputConversionBufferList = NULL;
        }
        if (audio->inputDeviceBufferList != NULL) {
            spDebug(100, "spCloseAudioDeviceArch", "call freeAudioBufferList for inputDeviceBufferList\n");
            freeAudioBufferList(audio->inputDeviceBufferList);
            audio->inputDeviceBufferList = NULL;
        }
        
        if (audio->inputUnit != NULL) {
            AudioUnitUninitialize(audio->inputUnit);
        }
        
        audio->inputUnitInitialized = SP_FALSE;
    }
    spUnlockMutex(audio->inputMutex);
        
    spLockMutex(audio->outputMutex);
    if (audio->outputUnitInitialized == SP_TRUE) {
        spDebug(100, "spCloseAudioDeviceArch", "free output buffers\n");
        freeInternalBuffer(audio, SP_TRUE);
        freeConversionBuffer(audio, SP_TRUE);
                
        if (audio->outputDeviceBufferList != NULL) {
            spDebug(100, "spCloseAudioDeviceArch", "call freeAudioBufferList for outputDeviceBufferList\n");
            freeAudioBufferList(audio->outputDeviceBufferList);
            audio->outputDeviceBufferList = NULL;
        }

        if (audio->outputUnit != NULL) {
            AudioUnitUninitialize(audio->outputUnit);
        }
        audio->outputUnitInitialized = SP_FALSE;
    }
    spUnlockMutex(audio->outputMutex);

    if (audio->inputConverter != NULL) {
        AudioConverterDispose(audio->inputConverter);
        audio->inputConverter = NULL;
    }
    if (audio->outputConverter != NULL) {
        AudioConverterDispose(audio->outputConverter);
        audio->outputConverter = NULL;
    }

    audio->inputEnabled = SP_FALSE;
    audio->outputEnabled = SP_FALSE;
    
    spDebug(100, "spCloseAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

long spReadAudioArch(spAudio audio, char *data, long length)
{
    int c;
    UInt32 k;
    UInt32 origSampByte;
    UInt32 sampByte;
    UInt32 copySize;
    UInt32 remainSize;
    UInt32 readLength;
    char *ptr;
    char *currentPtr;
    char *bufferPtr;
    short *sdata;
    short *sdata2;
    long *ldata;
    long *ldata2;
    spLong32 *l32data;
    UInt32 copyLength;
    
    spDebug(100, "spReadAudioArch", "in : length = %ld\n", length);
    
    if (audio->inputUnitInitialized == SP_FALSE) {
        if (openAudio(audio) == SP_FALSE) {
            return -1;
        }
    }
        
    origSampByte = spGetBitByte(audio->samp_bit, NULL);
    sampByte = (audio->samp_bit <= 32 && origSampByte >= 8) ? 4 : origSampByte;
    remainSize = (length / audio->num_channel) * sampByte;
    ptr = data;

    while (remainSize > 0) {
        if (audio->inputUnitStarted == SP_FALSE) {
            if (startInputAudio(audio) == SP_FALSE) {
                return -1;
            }
        }
        
        spDebug(100, "spReadAudioArch", "remainSize = %d, callbackCount = %d\n",
                remainSize, audio->inputCallbackCount);
#if 0
        while (audio->inputBufferFilled < audio->bufferSizeForInput) {
            spDebug(100, "spReadAudioArch", "inputBufferFilled = %d / %d\n",
                    audio->inputBufferFilled, audio->bufferSizeForInput);
            spWaitEventTimeout(audio->inputEvent, 1);
        }
#else
        spWaitEvent(audio->inputEvent);
#endif

        spDebug(100, "spReadAudioArch", "wait done: inputBufferFilled = %d / %d, inputBufferReadPos = %d / %d\n",
                audio->inputBufferFilled, audio->bufferSizeForInput, audio->inputBufferReadPos, audio->inputBufferSize);
        
        spLockMutex(audio->inputMutex);
        
        copySize = (audio->inputBufferSize - audio->inputBufferReadPos) % audio->bufferSizeForInput;
        if (copySize == 0) copySize = audio->bufferSizeForInput;
        copySize = MIN(copySize, audio->inputBufferFilled);
        copySize = MIN(copySize, remainSize);
        spDebug(100, "spReadAudioArch", "copySize = %d\n", copySize);
        
        if (audio->num_channel == 1) {
            if (audio->samp_bit <= 32 && origSampByte >= 8) {
                ldata = (long *)ptr;
                l32data = (spLong32 *)(audio->inputBuffer + audio->inputBufferReadPos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    if (audio->samp_bit == 24) {
                        ldata[k] = (long)(l32data[k] >> 8);
                    } else {
                        ldata[k] = (long)l32data[k];
                    }
                }
            } else if (audio->samp_bit == 24) {
                ldata = (long *)ptr;
                ldata2 = (long *)(audio->inputBuffer + audio->inputBufferReadPos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    ldata[k] = ldata2[k] >> 8;
                }
            } else if (audio->samp_bit == 8) {
                sdata = (short *)ptr;
                sdata2 = (short *)(audio->inputBuffer + audio->inputBufferReadPos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    sdata[k] = sdata2[k] >> 8;
                }
            } else {
                memcpy(ptr, audio->inputBuffer + audio->inputBufferReadPos, copySize);
            }
        } else {
            copyLength = copySize / sampByte;
            for (c = 0; c < audio->num_channel; c++) {
                for (k = 0; k < copyLength; k++) {
                    currentPtr = ptr + (k * audio->num_channel + c) * origSampByte;
                    bufferPtr = audio->inputBuffer + c * audio->inputBufferSize + audio->inputBufferReadPos + k * sampByte;
                    if (audio->samp_bit <= 32 && origSampByte >= 8) {
                        ldata = (long *)currentPtr;
                        l32data = (spLong32 *)bufferPtr;
                        if (audio->samp_bit == 24) {
                            *ldata = (long)*l32data >> 8;
                        } else {
                            *ldata = (long)*l32data;
                        }
                    } else if (audio->samp_bit == 24) {
                        ldata = (long *)currentPtr;
                        ldata2 = (long *)bufferPtr;
                        *ldata = *ldata2 >> 8;
                    } else if (audio->samp_bit == 8) {
                        sdata = (short *)currentPtr;
                        sdata2 = (short *)bufferPtr;
                        *sdata = *sdata2 >> 8;
                    } else {
                        memcpy(currentPtr, bufferPtr, sampByte);
                    }
                    
                }
            }
        }

        audio->inputBufferFilled -= copySize;
        audio->inputBufferReadPos += copySize;
        if (audio->inputBufferReadPos >= audio->inputBufferSize) {
            spDebug(100, "spReadAudioArch", "reset: inputBufferReadPos = %d\n", audio->inputBufferReadPos);
            audio->inputBufferReadPos = 0;
        }
        if (audio->inputBufferFilled < audio->bufferSizeForInput) {
            spDebug(100, "spReadAudioArch", "spResetEvent: inputBufferFilled = %d\n", audio->inputBufferFilled);
            spResetEvent(audio->inputEvent);
        }
        
        spUnlockMutex(audio->inputMutex);

        remainSize -= copySize;
        if (audio->samp_bit <= 32 && origSampByte >= 8) {
            ptr += audio->num_channel * copySize * 2;
        } else {
            ptr += audio->num_channel * copySize;
        }
        
        if (audio->block_mode == SP_AUDIO_NONBLOCKING
            && audio->inputBufferFilled < audio->bufferSizeForInput) {
            break;
        }
    }

    readLength = length - (audio->num_channel * remainSize / sampByte);
    
    spDebug(100, "spReadAudioArch", "done: remainSize = %d\n", remainSize);
    
    return readLength;
}

long spWriteAudioArch(spAudio audio, char *data, long length)
{
    int c;
    UInt32 k;
    UInt32 origSampByte;
    UInt32 sampByte;
    UInt32 copySize;
    UInt32 remainSize;
    char *ptr;
    char *currentPtr;
    char *bufferPtr;
    short *sdata;
    short *sdata2;
    long *ldata;
    long *ldata2;
    spLong32 *l32data;
    UInt32 copyLength;
    
    if (audio->outputUnitInitialized == SP_FALSE) {
        if (openAudio(audio) == SP_FALSE) {
            return -1;
        }
    }
        
    origSampByte = spGetBitByte(audio->samp_bit, NULL);
    sampByte = (audio->samp_bit <= 32 && origSampByte >= 8) ? 4 : origSampByte;
    remainSize = (length / audio->num_channel) * sampByte;
    spDebug(100, "spWriteAudioArch", "length = %ld, remainSize = %d\n", length, remainSize);
        
    ptr = data;

    while (remainSize > 0) {
        copySize = (audio->outputBufferSize - audio->outputBufferWritePos) % audio->bufferSizeForOutput;
        if (copySize == 0) copySize = audio->bufferSizeForOutput;
        copySize = MIN(copySize, remainSize);
        spDebug(100, "spWriteAudioArch", "outputBufferSize = %d, outputBufferWritePos = %d, copySize = %d\n",
                audio->outputBufferSize, audio->outputBufferWritePos, copySize);
        
        if (audio->outputBufferFilled + copySize > audio->outputBufferSize) {
            spDebug(100, "spWriteAudioArch", "spResetEvent: outputBufferFilled = %d / %d\n",
                    audio->outputBufferFilled, audio->outputBufferSize);
            spResetEvent(audio->outputEvent);
            spWaitEvent(audio->outputEvent);
            spDebug(100, "spWriteAudioArch", "spWaitEvent done\n");
        }
        
        spDebug(100, "spWriteAudioArch", "remainSize = %d, callbackCount = %d, outputBufferFilled = %d\n",
                remainSize, audio->outputCallbackCount, audio->outputBufferFilled);

        spLockMutex(audio->outputMutex);

        spDebug(100, "spWriteAudioArch", "memcpy: outputBufferWritePos = %d / %d, copySize = %d\n",
                audio->outputBufferWritePos, audio->outputBufferSize, copySize);

        if (audio->num_channel == 1) {
            if (audio->samp_bit <= 32 && origSampByte >= 8) {
                ldata = (long *)ptr;
                l32data = (spLong32 *)(audio->outputBuffer + audio->outputBufferWritePos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    if (audio->samp_bit == 24) {
                        l32data[k] = (spLong32)(0xffffff00 & (ldata[k] << 8));
                    } else {
                        l32data[k] = (spLong32)(0xffffffff & ldata[k]);
                    }
                }
            } else if (audio->samp_bit == 24) {
                ldata = (long *)ptr;
                ldata2 = (long *)(audio->outputBuffer + audio->outputBufferWritePos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    ldata2[k] = (0xffffff00 & (ldata[k] << 8));
                }
            } else if (audio->samp_bit == 8) {
                sdata = (short *)ptr;
                sdata2 = (short *)(audio->outputBuffer + audio->outputBufferWritePos);
                copyLength = copySize / sampByte;
                for (k = 0; k < copyLength; k++) {
                    sdata2[k] = (0xff00 & (sdata[k] << 8));
                }
            } else {
                memcpy(audio->outputBuffer + audio->outputBufferWritePos, ptr, copySize);
            }
        } else {
            copyLength = copySize / sampByte;
            for (c = 0; c < audio->num_channel; c++) {
                for (k = 0; k < copyLength; k++) {
                    currentPtr = ptr + (k * audio->num_channel + c) * origSampByte;
                    bufferPtr = audio->outputBuffer + c * audio->outputBufferSize + audio->outputBufferWritePos + k * sampByte;
                    if (audio->samp_bit <= 32 && origSampByte >= 8) {
                        ldata = (long *)currentPtr;
                        l32data = (spLong32 *)bufferPtr;
                        if (audio->samp_bit == 24) {
                            *l32data = (spLong32)(0xffffff00 & *ldata << 8);
                        } else {
                            *l32data = (spLong32)(0xffffffff & *ldata);
                        }
                    } else if (audio->samp_bit == 24) {
                        ldata = (long *)currentPtr;
                        ldata2 = (long *)bufferPtr;
                        *ldata2 = (0xffffff00 & (*ldata << 8));
                    } else if (audio->samp_bit == 8) {
                        sdata = (short *)currentPtr;
                        sdata2 = (short *)bufferPtr;
                        *sdata2 = (0xff00 & (*sdata << 8));
                    } else {
                        memcpy(bufferPtr, currentPtr, sampByte);
                    }
                }
            }
        }

        audio->outputBufferFilled += copySize;
        audio->outputBufferWritePos += copySize;
        if (audio->outputBufferWritePos >= audio->outputBufferSize) {
            spDebug(100, "spWriteAudioArch", "reset: outputBufferWritePos = %d\n", audio->outputBufferWritePos);
            audio->outputBufferWritePos = 0;
        }
        spUnlockMutex(audio->outputMutex);
        
        spDebug(100, "spWriteAudioArch", "outputBufferFilled = %d, bufferSizeForOutput = %d\n",
                audio->outputBufferFilled, audio->bufferSizeForOutput);
        
        if (audio->outputBufferFilled >= audio->bufferSizeForOutput) {
            if (audio->outputUnitStarted == SP_FALSE) {
                if (startOutputAudio(audio) == SP_FALSE) {
                    return -1;
                }
            }

            if (emitOutputPositionCallback(audio) == SP_FALSE) {
                break;
            }
        }
        
        remainSize -= copySize;
        if (audio->samp_bit <= 32 && origSampByte >= 8) {
            ptr += audio->num_channel * copySize * 2;
        } else {
            ptr += audio->num_channel * copySize;
        }
    }
    
    spDebug(100, "spWriteAudioArch", "done: remainSize = %ld\n", remainSize);
    
    return length;
}

spBool spGetAudioOutputPositionArch(spAudio audio, spLong *position)
{
    if (position != NULL) {
        if (audio->outputInitialSampleTime < 0.0) {
            *position = 0;
        } else {
            *position = (spLong)spRound((audio->outputSampleTime - audio->outputInitialSampleTime)
                                        * (audio->currentFormat.mSampleRate / audio->systemFormat.mSampleRate));
        }
        spDebug(80, "spGetAudioOutputPositionArch", "position = %ld, audio->outputBufferFilled = %d\n", *position, audio->outputBufferFilled);
    }
    return SP_TRUE;
}

spBool spStopAudioArch(spAudio audio)
{
    audio->doStop = SP_TRUE;
    
    if (audio->outputUnitStarted == SP_TRUE) {
        audio->outputUnitStarted = SP_FALSE;
        AudioOutputUnitStop(audio->outputUnit);
        spSetEvent(audio->outputEvent);
    }

    if (audio->inputUnitStarted == SP_TRUE) {
        audio->inputUnitStarted = SP_FALSE;
        AudioOutputUnitStop(audio->inputUnit);
        spSetEvent(audio->inputEvent);
    }
    
    return SP_TRUE;
}

spBool spSyncAudioArch(spAudio audio)
{
    if (audio->outputUnitStarted == SP_TRUE) {
        if (audio->outputEnabled == SP_TRUE) {
            waitOutputAudio(audio);
        }
    }
    
    resetBufferCounter(audio);
    
    audio->doStop = SP_FALSE;
    
    return SP_TRUE;
}

spBool spFreeAudioDriverArch(spAudio audio)
{
    if (audio->component != NULL) {
        if (audio->inputUnit != NULL) {
#if defined(IPHONE) | MAC_OS_X_VERSION_MAX_ALLOWED >= 1060
            AudioComponentInstanceDispose(audio->inputUnit);
#else
            CloseComponent(audio->inputUnit);
#endif
            audio->inputUnit = NULL;
            spDebug(80, "spFreeAudioDriverArch", "close input unit done\n");
        }
        if (audio->outputUnit != NULL) {
#if defined(IPHONE) | MAC_OS_X_VERSION_MAX_ALLOWED >= 1060
            AudioComponentInstanceDispose(audio->outputUnit);
#else
            CloseComponent(audio->outputUnit);
#endif
            audio->outputUnit = NULL;
            spDebug(80, "spFreeAudioDriverArch", "close output unit done\n");
        }
    }

    if (audio->inputEvent != NULL) {
        spDestroyEvent(audio->inputEvent);
    }
    if (audio->outputEvent != NULL) {
        spDestroyEvent(audio->outputEvent);
    }
    if (audio->inputMutex != NULL) {
        spDestroyMutex(audio->inputMutex);
    }
    if (audio->outputMutex != NULL) {
        spDestroyMutex(audio->outputMutex);
    }
    
    spDebug(80, "spFreeAudioDriverArch", "done\n");
    
    return SP_TRUE;
}

void spTerminateAudioArch(void)
{
    return;
}
#endif
