/*
 *        spAudio_WinArch.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spThread.h>

#include <sp/spWaveP.h>
#include <sp/spAudioP.h>

#if defined(SP_SUPPORT_AUDIO) && defined(_WIN32)
#include <stdio.h>
#include <stdlib.h>

#include "spAudio_WinArch.h"

#define SP_MAX_NUM_WAVE_BUFFER 256
#define SP_NUM_WAVE_PREPARE_BUFFER 8

struct _spWinAudio {
    int num_channel;
    int samp_byte;
    long buffer_size;
    long num_prepare_buffer;
    spBool header_allocated;
    spBool start_flag;
    spBool wait_flag;
    spBool nonblock_flag;

    long buffer_index;
    long prev_index;
    long buffer_pos;
    unsigned long prev_sample;
    
    long num_buffer;
    HANDLE hbuffer[SP_MAX_NUM_WAVE_BUFFER];
    HPSTR lpbuffer[SP_MAX_NUM_WAVE_BUFFER];
    HGLOBAL hwavehdr[SP_MAX_NUM_WAVE_BUFFER];
    LPWAVEHDR lpwavehdr[SP_MAX_NUM_WAVE_BUFFER];
    CRITICAL_SECTION critical_section;
    spBool critical_section_initialized;

    HANDLE hevent;
    long process_buffer_index;
    long num_process_buffer;
    
    HWAVEIN hwavein;
    HWAVEOUT hwaveout;
    WAVEFORMATEX waveformat;

    spAudioCallbackType call_type;
    spAudioCallbackFunc_Win call_func;
    void *call_data;
};

static spBool resetBufferEvent(spWinAudio audio)
{
    spBool flag = SP_FALSE;
    
    EnterCriticalSection(&audio->critical_section);
    if (audio->num_process_buffer > 0) {
        flag = SP_TRUE;
        ResetEvent(audio->hevent);
    } else {
        SetEvent(audio->hevent);
    }
    LeaveCriticalSection(&audio->critical_section);
    
    return flag;
}

static void incrementBufferCounter(spWinAudio audio, long *increment_counter, long *increment_counter_nolimit,
                                   long *decrement_counter)
{
    EnterCriticalSection(&audio->critical_section);

    if (increment_counter != NULL) {
        (*increment_counter)++;
        if (*increment_counter >= audio->num_buffer) *increment_counter = 0;
        spDebug(100, "incrementBufferCounter", "updated: increment_counter = %ld\n", *increment_counter);
    }
    if (increment_counter_nolimit != NULL) {
        (*increment_counter_nolimit)++;
        spDebug(100, "incrementBufferCounter", "updated: increment_counter_nolimit = %ld\n",
                *increment_counter_nolimit);
    }
    if (decrement_counter != NULL) {
        (*decrement_counter)--;
        if (*decrement_counter < 0) *decrement_counter = 0;
        spDebug(100, "incrementBufferCounter", "updated: decrement_counter = %ld\n", *decrement_counter);
    }
    
    LeaveCriticalSection(&audio->critical_section);
    
    return;
}

spAudioCallbackType spSetAudioCallbackFunc_Win(spWinAudio audio, spAudioCallbackType call_type,
                                               spAudioCallbackFunc_Win call_func, void *call_data)
{
    audio->call_type = call_type;
    audio->call_func = call_func;
    audio->call_data = call_data;

    return audio->call_type;
}

static void freeWaveBuffer(spWinAudio audio)
{
    long i;

    if (audio->num_buffer > 0) {
        for (i = 0; i < audio->num_buffer; i++) {
            if (audio->lpbuffer[i] != NULL) {
                GlobalUnlock(audio->hbuffer[i]);
                audio->lpbuffer[i] = NULL;
            }
            if (audio->hbuffer[i] != NULL) {
                GlobalFree(audio->hbuffer[i]);
                audio->hbuffer[i] = NULL;
            }
        }
    }
    
    if (audio->hevent != NULL) {
        CloseHandle(audio->hevent);
    }
    if (audio->critical_section_initialized == SP_TRUE) {
        DeleteCriticalSection(&audio->critical_section);
    }
    
    audio->num_buffer = 0;
    
    return;
}

static spBool allocWaveBuffer(spWinAudio audio, long num_buffer, long buffer_size)
{
    long i;

    if (audio->num_buffer > 0) {
        for (i = 0; i < audio->num_buffer; i++) {
            memset(audio->lpbuffer[i], 0, audio->buffer_size);
        }
        
        return SP_TRUE;
    }
    
    InitializeCriticalSection(&audio->critical_section);
    audio->critical_section_initialized = SP_TRUE;
    
    for (i = 0; i < num_buffer; i++) {
        /* allocate and lock memory for the waveform buffer */ 
        if ((audio->hbuffer[i] = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT,
                                             buffer_size)) == NULL) {
            spMessage("can't allocate data memory\n");
            freeWaveBuffer(audio);
            return SP_FALSE;
        }

        if ((audio->lpbuffer[i] = GlobalLock(audio->hbuffer[i])) == NULL) {
            spMessage("can't lock data memory\n");
            freeWaveBuffer(audio);
            return SP_FALSE;
        }
    }
    
    audio->num_buffer = num_buffer;
    audio->buffer_size = buffer_size;

    audio->hevent = CreateEvent(NULL, TRUE, TRUE, NULL);

    return SP_TRUE;
}

static void freeWaveHdr(spWinAudio audio)
{
    long i;

    if (audio->header_allocated == SP_TRUE) {
        for (i = 0; i < audio->num_buffer; i++) {
            if (audio->lpwavehdr[i] != NULL) {
                GlobalUnlock(audio->hwavehdr[i]);
                audio->lpwavehdr[i] = NULL;
            }
            if (audio->hwavehdr[i] != NULL) {
                GlobalFree(audio->hwavehdr[i]);
                audio->hwavehdr[i] = NULL;
            }
        }
    }
    
    audio->header_allocated = SP_FALSE;
    
    return;
}

static spBool allocWaveHdr(spWinAudio audio)
{
    long i;
    
    if (audio->header_allocated == SP_TRUE) {
        return SP_TRUE;
    }
        
    for (i = 0; i < audio->num_buffer; i++) {
        if ((audio->hwavehdr[i] = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT,
                                              sizeof(WAVEHDR))) == NULL) {
            spMessage("can't allocate header memory\n");
            freeWaveHdr(audio);
            freeWaveBuffer(audio);
            return SP_FALSE;
        }
        
        if ((audio->lpwavehdr[i] = GlobalLock(audio->hwavehdr[i])) == NULL) {
            spMessage("can't lock header memory\n");
            freeWaveHdr(audio);
            freeWaveBuffer(audio);
            return SP_FALSE;
        }
    }

    audio->header_allocated = SP_TRUE;
    
    return SP_TRUE;
}

static void waveProcMain(spWinAudio audio, spBool output_flag)
{
    if (audio == NULL) return;

    spDebug(100, "waveProcMain", "output_flag = %d\n", output_flag);
    
    EnterCriticalSection(&audio->critical_section);
        
    audio->process_buffer_index++;
    if (audio->process_buffer_index >= audio->num_buffer) audio->process_buffer_index = 0;

    audio->num_process_buffer--;
    if (audio->num_process_buffer < 0) audio->num_process_buffer = 0;
        
    spDebug(100, "waveProcMain",
            "process_buffer_index = %ld, num_process_buffer = %ld, wait_flag = %d\n",
            audio->process_buffer_index, audio->num_process_buffer, audio->wait_flag);
        
    SetEvent(audio->hevent);
    
    LeaveCriticalSection(&audio->critical_section);

    spDebug(100, "waveProcMain", "done\n");
        
    return;
}

static void CALLBACK waveInProc(HWAVEIN hwavein, UINT uMsg, DWORD_PTR dwInstance,  
                                 DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
    spWinAudio audio = (spWinAudio)dwInstance;

    switch (uMsg) {
      case WIM_DATA:
        waveProcMain(audio, SP_FALSE);
        break;
      default:
        break;
    }

    return;
}

static void CALLBACK waveOutProc(HWAVEOUT hwaveout, UINT uMsg, DWORD_PTR dwInstance,  
                                 DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
    long length;
    WAVEHDR *wavehdr;
    spWinAudio audio = (spWinAudio)dwInstance;
    
    switch (uMsg) {
      case WOM_DONE:
        waveProcMain(audio, SP_TRUE);
        
        if (audio->call_func != NULL
            && (audio->call_type & SP_AUDIO_OUTPUT_BUFFER_CALLBACK)) {
            wavehdr = (WAVEHDR *)dwParam1;
            length = wavehdr->dwBufferLength / audio->samp_byte;
            audio->call_func(audio, SP_AUDIO_OUTPUT_BUFFER_CALLBACK,
                             wavehdr->lpData, &length, audio->call_data);
        }
        break;
      default:
        break;
    }

    return;
}

static spBool openDevice(spWinAudio audio, UINT device_id, int samp_bit, int num_channel,
                         double samp_rate, spBool nonblock_flag, spBool output_flag)
{
    int err;
    WAVEFORMATEX waveFormatEx;
    DWORD_PTR dwCallback;
    DWORD fdwOpen;
        
    waveFormatEx.wFormatTag = WAVE_FORMAT_PCM;
    waveFormatEx.nChannels = num_channel;
    waveFormatEx.nSamplesPerSec = (long)spRound(samp_rate);
    waveFormatEx.wBitsPerSample = samp_bit;
    waveFormatEx.nBlockAlign = waveFormatEx.nChannels * waveFormatEx.wBitsPerSample / 8;
    waveFormatEx.nAvgBytesPerSec =
        waveFormatEx.nSamplesPerSec * waveFormatEx.nBlockAlign;
    waveFormatEx.cbSize = 0;

    if (output_flag == SP_TRUE) {
        dwCallback = (DWORD_PTR)waveOutProc;
    } else {
        dwCallback = (DWORD_PTR)waveInProc;
    }
    fdwOpen = CALLBACK_FUNCTION;
    
    if (output_flag == SP_TRUE) {
        /*fdwOpen |= WAVE_ALLOWSYNC;*/
        
        /* open a waveform output device */ 
        if ((err = waveOutOpen(&audio->hwaveout,                 /*LPHWAVEOUT phwo*/
                               device_id,                         /*UINT uDeviceID*/
                               (LPWAVEFORMATEX)&waveFormatEx,         /*LPWAVEFORMATEX pwfx*/
                               (DWORD_PTR)dwCallback,                 /*DWORD_PTR dwCallback*/
                               (DWORD_PTR)audio,                 /*DWORD_PTR dwCallbackInstance*/
                               fdwOpen                                 /*DWORD fdwOpen*/
                               )) != MMSYSERR_NOERROR) {
            spMessage("open error: %d\n", err);
            return SP_FALSE;
        }
        audio->hwavein = NULL;
    } else {
        /* open a waveform input device */ 
        if ((err = waveInOpen(&audio->hwavein,                         /*LPHWAVEIN phwi*/
                              device_id,                         /*UINT uDeviceID*/
                              (LPWAVEFORMATEX)&waveFormatEx,         /*LPWAVEFORMATEX pwfx*/
                              (DWORD_PTR)dwCallback,                 /*DWORD_PTR dwCallback*/
                              (DWORD_PTR)audio,                         /*DWORD_PTR dwCallbackInstance*/
                              fdwOpen                                /*DWORD fdwOpen*/
                              )) != MMSYSERR_NOERROR) {
            spMessage("open error: %d\n", err);
            return SP_FALSE;
        }
        audio->hwaveout = NULL;
    }

    audio->nonblock_flag = nonblock_flag;
    audio->num_channel = waveFormatEx.nChannels;
    audio->samp_byte = waveFormatEx.wBitsPerSample / 8;
    spDebug(50, "openDevice", "num_channel = %d, samp_byte = %d\n", audio->num_channel, audio->samp_byte);

    return SP_TRUE;
}

static spBool waitBuffer(spWinAudio audio, long index)
{
    spDebug(100, "waitBuffer", "index = %ld, process_buffer_index = %ld, num_process_buffer = %ld\n",
            index, audio->process_buffer_index, audio->num_process_buffer);
    if (index >= 0) {
        if (audio->process_buffer_index == index) {
            if (resetBufferEvent(audio) == SP_TRUE) {
                WaitForSingleObject(audio->hevent, INFINITE);
            }
        }
    }
    return SP_TRUE;
}

static spBool emitOutputPositionCallback(spWinAudio audio)
{
    spBool noerr = SP_TRUE;
    
    spDebug(1, "emitOutputPositionCallback", "audio->call_type = %ld\n", audio->call_type);
    
    if (audio->call_func != NULL) {
        if (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK) {
            spLong position;

            if (spGetAudioOutputPosition_Win(audio, &position) == SP_TRUE) {
                if (audio->call_func(audio, SP_AUDIO_OUTPUT_POSITION_CALLBACK,
                                     &position, NULL, audio->call_data) == SP_FALSE) {
                    noerr = SP_FALSE;
                }
            }
        }
    }

    if (noerr == SP_FALSE) {
        audio->start_flag = SP_FALSE;
    }
        
    return noerr;
}

static spBool waitFinish(spWinAudio audio)
{
    if (audio->start_flag == SP_TRUE) {
        EnterCriticalSection(&audio->critical_section);
        audio->wait_flag = SP_TRUE;
        LeaveCriticalSection(&audio->critical_section);

        while (1) {
            if (audio->wait_flag == SP_FALSE || audio->start_flag == SP_FALSE) {
                break;
            }
            if (emitOutputPositionCallback(audio) == SP_FALSE) {
                break;
            }
            
            if (resetBufferEvent(audio) == SP_FALSE) {
                break;
            }
            WaitForSingleObject(audio->hevent, INFINITE);
        }
    }
    
    EnterCriticalSection(&audio->critical_section);
    audio->wait_flag = SP_FALSE;
    SetEvent(audio->hevent);
    LeaveCriticalSection(&audio->critical_section);
    
    return SP_TRUE;
}

static spBool readWaveFrame(spWinAudio audio)
{
    int err;
    
    spDebug(100, "readWaveFrame", "in\n");
    
    incrementBufferCounter(audio, &audio->buffer_index, &audio->num_process_buffer, NULL);
    
    if ((err = waveInAddBuffer(audio->hwavein, audio->lpwavehdr[audio->buffer_index],
                               sizeof(WAVEHDR))) != MMSYSERR_NOERROR) {
        incrementBufferCounter(audio, NULL, NULL, &audio->num_process_buffer);
        spMessage("read error: %d\n", err);
        return SP_FALSE;
    }
    
    spDebug(100, "readWaveFrame", "done\n");
    
    return SP_TRUE;
}

static spBool writeWaveFrame(spWinAudio audio)
{
    int err;
    
    spDebug(100, "writeWaveFrame", "in\n");
    
    incrementBufferCounter(audio, NULL, &audio->num_process_buffer, NULL);
    
    if ((err = waveOutWrite(audio->hwaveout, audio->lpwavehdr[audio->buffer_index],
                            sizeof(WAVEHDR))) != MMSYSERR_NOERROR) {
        incrementBufferCounter(audio, NULL, NULL, &audio->num_process_buffer);
        spMessage("write error: %d\n", err);
        return SP_FALSE;
    }
    
    audio->prev_index = audio->buffer_index;
    incrementBufferCounter(audio, &audio->buffer_index, NULL, NULL);
    
    spDebug(100, "writeWaveFrame", "done\n");
    
    return SP_TRUE;
}

static spBool prepareWaveHdr(spWinAudio audio)
{
    long i;

    spDebug(100, "prepareWaveHdr", "in\n");
    
    for (i = 0; i < audio->num_buffer; i++) {
        audio->lpwavehdr[i]->lpData = audio->lpbuffer[i];
        audio->lpwavehdr[i]->dwBufferLength = audio->buffer_size;
        audio->lpwavehdr[i]->dwFlags = 0;
        audio->lpwavehdr[i]->dwLoops = 0;

        if (audio->hwavein != NULL) {
            if (waveInPrepareHeader(audio->hwavein, audio->lpwavehdr[i], sizeof(WAVEHDR))
                != MMSYSERR_NOERROR) {
                spMessage("buffer preparation error.\n");
                return SP_FALSE;
            }
        } else {
            if (waveOutPrepareHeader(audio->hwaveout, audio->lpwavehdr[i], sizeof(WAVEHDR))
                != MMSYSERR_NOERROR) {
                spMessage("buffer preparation error.\n");
                return SP_FALSE;
            }
        }
        
        if (!(audio->lpwavehdr[i]->dwFlags & WHDR_PREPARED)) {
            spMessage("buffer preparation error.\n");
            return SP_FALSE;
        }
    }
    spResetDevice_Win(audio);

    if (audio->hwavein != NULL) {
        audio->buffer_index = -1;
        for (i = 0; i < audio->num_buffer; i++) {
            if (readWaveFrame(audio) == SP_FALSE) {
                return SP_FALSE;
            }
        }
    } else {
        audio->buffer_index = 0;
        for (i = 0; i < audio->num_prepare_buffer; i++) {
            if (writeWaveFrame(audio) == SP_FALSE) {
                return SP_FALSE;
            }
        }
    }
    
    spDebug(100, "prepareWaveHdr", "done\n");
    
    return SP_TRUE;
}

static spBool unprepareWaveHdr(spWinAudio audio)
{
    long i;

    if (audio->hwavein != NULL) {
        waveInReset(audio->hwavein);
        
        for (i = 0; i < audio->num_buffer; i++) {
            if (waveInUnprepareHeader(audio->hwavein, audio->lpwavehdr[i], sizeof(WAVEHDR))
                != MMSYSERR_NOERROR) {
                spMessage("buffer unpreparation error.\n");
                return SP_FALSE;
            }
        }
    } else {
        waitFinish(audio);
        waveOutReset(audio->hwaveout);
        
        for (i = 0; i < audio->num_buffer; i++) {
            if (waveOutUnprepareHeader(audio->hwaveout, audio->lpwavehdr[i], sizeof(WAVEHDR))
                != MMSYSERR_NOERROR) {
                spMessage("buffer unpreparation error.\n");
                return SP_FALSE;
            }
        }
    }
        
    audio->prev_index = -1;
    
    return SP_TRUE;
}

static spBool closeDevice(spWinAudio audio)
{
    if (audio->hwavein != NULL) {
        waveInClose(audio->hwavein);
    } else {
        waveOutClose(audio->hwaveout);
    }
    audio->start_flag = SP_FALSE;

    return SP_TRUE;
}

spWinAudio spInitAudio_Win(long num_buffer, long buffer_size)
{
    spWinAudio audio;

    spDebug(1, "spInitAudio_Win", "num_buffer = %ld, buffer_size = %ld\n",
            num_buffer, buffer_size);
    
    num_buffer = MIN(num_buffer, SP_MAX_NUM_WAVE_BUFFER);

    audio = xalloc(1, struct _spWinAudio);
    memset(audio, 0, sizeof(struct _spWinAudio));
    audio->num_prepare_buffer = MIN(SP_NUM_WAVE_PREPARE_BUFFER, num_buffer/2);
    audio->num_channel = 1;
    audio->prev_index = -1;
    audio->critical_section_initialized = SP_FALSE;

    if (allocWaveBuffer(audio, MIN(num_buffer, SP_MAX_NUM_WAVE_BUFFER), buffer_size) == SP_FALSE) {
        xfree(audio);
        spDebug(10, "spInitAudio_Win", "allocWaveBuffer failed\n");
        return NULL;
    }
    if (allocWaveHdr(audio) == SP_FALSE) {
        xfree(audio);
        spDebug(10, "spInitAudio_Win", "allocWaveHdr failed\n");
        return NULL;
    }

    spDebug(50, "spInitAudio_Win", "done\n");
    
    return audio;
}

spBool spFreeAudio_Win(spWinAudio audio)
{
    if (audio == NULL) return SP_FALSE;

    freeWaveHdr(audio);
    freeWaveBuffer(audio);

    return SP_TRUE;
}

spBool spOpenAudioDevice_Win(spWinAudio audio, UINT device_id, int samp_bit, int num_channel,
                             double samp_rate, spBool nonblock_flag, spBool output_flag)
{
    if (audio == NULL) return SP_FALSE;
    
    spDebug(100, "spOpenAudioDevice_Win", "output_flag = %d\n", output_flag);
    
    if (openDevice(audio, device_id, samp_bit, num_channel, samp_rate, nonblock_flag, output_flag) == SP_FALSE) {
        spDebug(10, "spOpenAudioDevice_Win", "openDevice failed\n");
        return SP_FALSE;
    }
    if (prepareWaveHdr(audio) == SP_FALSE) {
        spDebug(10, "spOpenAudioDevice_Win", "prepareWaveHdr failed\n");
        return SP_FALSE;
    }

    spDebug(100, "spOpenAudioDevice_Win", "done\n");
    
    return SP_TRUE;
}

spBool spCloseAudioDevice_Win(spWinAudio audio)
{
    if (audio == NULL) return SP_FALSE;
    
    unprepareWaveHdr(audio);
    closeDevice(audio);
    
    return SP_TRUE;
}

long spReadWave_Win(spWinAudio audio, char *data, long length)
{
    long index;
    long count;
    long ncopy;
    long nremain;

    if (audio->hwavein == NULL) return -1;

    if (audio->start_flag == SP_FALSE) {
        waveInStart(audio->hwavein);
        audio->start_flag = SP_TRUE;
        audio->prev_sample = 0;
    }

    if (audio->nonblock_flag == SP_TRUE) {
        MMTIME mmtime;
        unsigned long current_sample;
        unsigned long current_length;
        
        mmtime.wType = TIME_SAMPLES;
        waveInGetPosition(audio->hwavein, &mmtime, sizeof(mmtime));

        current_sample = mmtime.u.sample * audio->num_channel;
        current_length = current_sample - audio->prev_sample;

        if (current_length <= 0) {
            current_length = audio->buffer_size / audio->samp_byte;
        }
        length = MIN(length, (long)current_length);
        spDebug(1, NULL, "position = %ld, length = %ld\n",
                current_sample, length);
        
        audio->prev_sample += length;
    }

    count = 0;
    nremain = length * audio->samp_byte;
    
    if (audio->prev_index >= 0 && audio->buffer_pos > 0) {
        ncopy = audio->buffer_size - audio->buffer_pos;
        if (nremain <= ncopy) {
            memmove(data, audio->lpbuffer[audio->prev_index] + audio->buffer_pos, nremain);
            count = nremain;
            audio->buffer_pos += nremain;
            nremain = 0;
        } else {
            if (ncopy > 0) {
                memmove(data, audio->lpbuffer[audio->prev_index] + audio->buffer_pos, ncopy);
                count = ncopy;
                nremain -= count;
            } else {
                count = 0;
            }
            audio->buffer_pos = 0;
            
            if (readWaveFrame(audio) == SP_FALSE) {
                return SP_FALSE;
            }
        }
    }

    if (nremain > 0) {
        while (1) {
            index = audio->buffer_index - audio->num_buffer + 1;
            if (index < 0) index += audio->num_buffer;
            spDebug(100, "spReadWave_Win", "index = %ld, nremain = %ld\n", index, nremain);

            waitBuffer(audio, index);
        
            audio->prev_index = index;
        
            if (nremain <= audio->buffer_size) {
                if (nremain > 0) {
                    memmove(data + count, audio->lpbuffer[index], nremain);
                    count += nremain;
                }
                audio->buffer_pos = nremain;
                nremain = 0;
                break;
            } else {
                memmove(data + count, audio->lpbuffer[index], audio->buffer_size);
                count += audio->buffer_size;
                nremain -= audio->buffer_size;
                audio->buffer_pos = 0;
            }
            
            if (readWaveFrame(audio) == SP_FALSE) {
                return SP_FALSE;
            }
        }
    }
    
    spDebug(100, "spReadWave_Win", "done: count = %ld\n", count);
    
    return count / audio->samp_byte;
}

long spWriteWave_Win(spWinAudio audio, char *data, long length)
{
    long count;
    long nremain;
    long ncopy;

    spDebug(100, "spWriteWave_Win", "in\n");
        
    if (audio->hwaveout == NULL) return -1;

    if (audio->start_flag == SP_FALSE) {
        audio->start_flag = SP_TRUE;
    }

    count = 0;
    nremain = length * audio->samp_byte;

    while (1) {
        spDebug(100, "spWriteWave_Win", "buffer_index = %ld, nremain = %ld\n",
                audio->buffer_index, nremain);
        
        waitBuffer(audio, audio->buffer_index);
        
        ncopy = MIN(nremain, audio->buffer_size);
        memmove(audio->lpbuffer[audio->buffer_index],
                data + count, ncopy);
        audio->lpwavehdr[audio->buffer_index]->dwBufferLength = ncopy;
        count += ncopy;
        nremain -= ncopy;
        
        if (writeWaveFrame(audio) == SP_FALSE) {
            return SP_FALSE;
        }
        
        if (audio->start_flag == SP_FALSE) {
            break;
        }
        if (emitOutputPositionCallback(audio) == SP_FALSE) {
            break;
        }
        
        if (nremain <= 0) {
            break;
        }
    }

    spDebug(100, "spWriteWave_Win", "done: count = %ld\n", count);
        
    return count / audio->samp_byte;
}

spBool spGetAudioOutputPosition_Win(spWinAudio audio, spLong *position)
{
    MMTIME mmtime;
    DWORD offset;

    if (audio == NULL) return SP_FALSE;

    if (position != NULL) {
        mmtime.wType = TIME_SAMPLES;
        if (waveOutGetPosition(audio->hwaveout, &mmtime, sizeof(mmtime)) != MMSYSERR_NOERROR) {
            return SP_FALSE;
        }

        offset = (DWORD)(audio->buffer_size / audio->samp_byte) * (DWORD)audio->num_prepare_buffer;
        spDebug(80, "spGetAudioOutputPosition_Win", "mmtime.u.sample = %ld, offset = %ld\n",
                mmtime.u.sample, offset);

        if (offset > mmtime.u.sample) {
            *position = 0;
        } else {
            *position = (spLong)(mmtime.u.sample - offset);
        }
    }
    
    return SP_TRUE;
}

spBool spResetDevice_Win(spWinAudio audio)
{
    if (audio == NULL) return SP_FALSE;
    
    spDebug(50, "spResetDevice_Win", "in\n");
    
    if (audio->hwavein != NULL) {
        waveInReset(audio->hwavein);
    } else if (audio->hwaveout != NULL) {
        waveOutReset(audio->hwaveout);
    } else {
        return SP_FALSE;
    }
    
    EnterCriticalSection(&audio->critical_section);
    audio->start_flag = SP_FALSE;
    audio->wait_flag = SP_FALSE;
    audio->prev_index = -1;
    audio->buffer_pos = 0;
    audio->process_buffer_index = 0;
    SetEvent(audio->hevent);
    LeaveCriticalSection(&audio->critical_section);
    
    spDebug(50, "spResetDevice_Win", "done\n");
    
    return SP_TRUE;
}

spBool spSyncDevice_Win(spWinAudio audio)
{
    long i;

    if (audio == NULL) return SP_FALSE;
    
    spDebug(50, "spSyncDevice_Win", "in\n");
    
    if (audio->hwavein != NULL) {
        spResetDevice_Win(audio);
        
        audio->buffer_index = -1;
        for (i = 0; i < audio->num_buffer; i++) {
            if (readWaveFrame(audio) == SP_FALSE) {
                return SP_FALSE;
            }
        }
    } else {
        waitFinish(audio);
        spResetDevice_Win(audio);
        
        if (audio->num_prepare_buffer > 0) {
            audio->buffer_index = 0;
            for (i = 0; i < audio->num_prepare_buffer; i++) {
                if (writeWaveFrame(audio) == SP_FALSE) {
                    return SP_FALSE;
                }
            }
        }
    }
    
    spDebug(50, "spSyncDevice_Win", "done\n");
    
    return SP_TRUE;
}

#endif

