/*
 *        spID3.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spString.h>
#include <sp/spKanji.h>

#include <sp/spChunkP.h>
#include <sp/spID3.h>

static char sp_id3_type_id_version_table[][3][5] = {
    {SP_ID3_FRAME_TYPE_ALBUM_V2_3, SP_ID3_FRAME_TYPE_ALBUM_V2_3, SP_ID3_FRAME_TYPE_ALBUM_V2_2},
    {SP_ID3_FRAME_TYPE_COMPOSER_V2_3, SP_ID3_FRAME_TYPE_COMPOSER_V2_3, SP_ID3_FRAME_TYPE_COMPOSER_V2_2},
    {SP_ID3_FRAME_TYPE_LYRICIST_V2_3, SP_ID3_FRAME_TYPE_LYRICIST_V2_3, SP_ID3_FRAME_TYPE_LYRICIST_V2_2},
    {SP_ID3_FRAME_TYPE_COPYRIGHT_V2_3, SP_ID3_FRAME_TYPE_COPYRIGHT_V2_3, SP_ID3_FRAME_TYPE_COPYRIGHT_V2_2},
    {SP_ID3_FRAME_TYPE_TRACK_V2_3, SP_ID3_FRAME_TYPE_TRACK_V2_3, SP_ID3_FRAME_TYPE_TRACK_V2_2},
    {SP_ID3_FRAME_TYPE_DISC_V2_3, SP_ID3_FRAME_TYPE_DISC_V2_3, SP_ID3_FRAME_TYPE_DISC_V2_2},
    {SP_ID3_FRAME_TYPE_TITLE_V2_3, SP_ID3_FRAME_TYPE_TITLE_V2_3, SP_ID3_FRAME_TYPE_TITLE_V2_2},
    {SP_ID3_FRAME_TYPE_ARTIST_V2_3, SP_ID3_FRAME_TYPE_ARTIST_V2_3, SP_ID3_FRAME_TYPE_ARTIST_V2_2},
    {SP_ID3_FRAME_TYPE_GENRE_V2_3, SP_ID3_FRAME_TYPE_GENRE_V2_3, SP_ID3_FRAME_TYPE_GENRE_V2_2},
    {SP_ID3_FRAME_TYPE_SOFTWARE_V2_3, SP_ID3_FRAME_TYPE_SOFTWARE_V2_3, SP_ID3_FRAME_TYPE_SOFTWARE_V2_2},
    {SP_ID3_FRAME_TYPE_TEMPO_V2_3, SP_ID3_FRAME_TYPE_TEMPO_V2_3, SP_ID3_FRAME_TYPE_TEMPO_V2_2},
    {SP_ID3_FRAME_TYPE_RELEASE_V2_4, SP_ID3_FRAME_TYPE_RELEASE_V2_3, SP_ID3_FRAME_TYPE_RELEASE_V2_2},
    {SP_ID3_FRAME_TYPE_ISRC_V2_3, SP_ID3_FRAME_TYPE_ISRC_V2_3, ""},
    {SP_ID3_FRAME_TYPE_PUBLISHER_V2_3, SP_ID3_FRAME_TYPE_PUBLISHER_V2_3, SP_ID3_FRAME_TYPE_PUBLISHER_V2_2},
    {"", "", ""},
};

const char *spGetID3CurrentVersionTypeId(spID3Header *header, const char *type)
{
    int i;
    int version_index;

    if (header->version >= 4) {
        version_index = 0;
    } else if (header->version == 3) {
        version_index = 1;
    } else {
        version_index = 2;
    }

    for (i = 0;; i++) {
        if (sp_id3_type_id_version_table[i][0][0] == NUL) {
            break;
        }
        spDebug(100, "spGetID3CurrentVersionTypeId", "i = %d: type = %s, entry = %s\n",
                i, type, sp_id3_type_id_version_table[i][version_index]);

        if (streq(sp_id3_type_id_version_table[i][0], type)
            || streq(sp_id3_type_id_version_table[i][1], type)
            || streq(sp_id3_type_id_version_table[i][2], type)) {
            if (sp_id3_type_id_version_table[i][version_index][0] == NUL) {
                return NULL;
            } else {
                return sp_id3_type_id_version_table[i][version_index];
            }
        }
    }

    return NULL;
}

spID3Header *spGetID3HeaderFromFrame(spID3Frame *frame)
{
    spID3Frame *parent;

    parent = frame;
    while (parent->parent != NULL) {
        parent = parent->parent;
    }

    return (spID3Header *)parent;
}

spBool spIsID3NullFrameHeader(spID3FrameHeader *frame_header)
{
    if (frame_header->type[0] == 0 && frame_header->type[1] == 0 && frame_header->type[2] == 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsID3FrameUnsynchronized(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version >= 4) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK) {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    } else {
        if (header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK) {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    }
}

void spSetID3FrameUnsynchronizedFlag(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version >= 4) {
        frame->header.flag2 |= SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK;
    } else {
        header->header.flag1 |= SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK;
    }

    return;
}

void spUpdateID3FrameUnsynchronizedFlag(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->unsync_count > 0) {
        if (header->version >= 4) {
            frame->header.flag2 |= SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK;
        } else {
            header->header.flag1 |= SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK;
        }
    }

    return;
}

long spReadID3SyncSafeLong32(unsigned long *value, FILE *fp)
{
    long nread;
    unsigned char buf[4];

    if ((nread = (long)fread(buf, 1, 4, fp)) != 4) {
        return nread;
    }

    *value = (0x7f & (unsigned long)buf[0]) << 21 | (0x7f & (unsigned long)buf[1]) << 14
        | (0x7f & (unsigned long)buf[2]) << 7 | (0x7f & (unsigned long)buf[3]);

    return nread;
}

long spWriteID3SyncSafeLong32(unsigned long value, FILE *fp)
{
    long nwrite;
    unsigned char buf[4];

    buf[0] = (unsigned char)(0x7f & (value >> 21));
    buf[1] = (unsigned char)(0x7f & (value >> 14));
    buf[2] = (unsigned char)(0x7f & (value >> 7));
    buf[3] = (unsigned char)(0x7f & value);

    if ((nwrite = (long)fwrite(buf, 1, 4, fp)) != 4) {
        return nwrite;
    }

    return nwrite;
}

long spReadID3UnsynchronizedBuffer(unsigned char *buf, long need_len, long read_len, long block_len, long stride,
                                   int swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long k, n;
    long ret;
    long nread;
    unsigned char prev_value;
    long shift;
    long swap_size = 0;

    prev_value = 0;
    if (last_value != NULL) prev_value = *last_value;

    if (swap_flag) {
        swap_size = block_len;
    }
    
    if (swap_size > 1) {
        shift = swap_size - 1;
    } else {
        shift = 0;
    }

    for (k = 0, nread = 0;;) {
        if (!SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == 0) {
                n = k;
                k += (stride - block_len);
                for (; n < k; n++) {
                    buf[k] = 0;
                }
            }
        }

        if ((ret = (long)fread(&buf[k + shift], 1, 1, fp)) != 1) {
            spDebug(10, "spReadID3UnsynchronizedBuffer", "fread failed: %ld\n", ret);
            return ret;
        }
        ++nread;

        if (unsync_flag) {
            if (prev_value == 0xff && buf[k + shift] == 0) {
                if ((ret = (long)fread(&buf[k + shift], 1, 1, fp)) != 1) {
                    spDebug(10, "spReadID3UnsynchronizedBuffer", "fread failed: %ld\n", ret);
                    return ret;
                }
                if (unsync_count != NULL) {
                    *unsync_count += 1;
                }
                spDebug(10, "spReadID3UnsynchronizedBuffer", "updated buf[%ld] = %x, nread = %ld\n", k, buf[k], nread);
                ++nread;
            }
        }

        prev_value = buf[k + shift];
        if (last_value != NULL) *last_value = prev_value;
        
        ++k;
        
        if (SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == block_len) {
                n = k;
                k += (stride - block_len);
                for (; n < k; n++) {
                    buf[k] = 0;
                }
            }
        }
        
        if (read_len > 0 && nread >= read_len) {
            break;
        }
        
        if (need_len > 0 && k >= need_len) {
            break;
        }
        
        if (swap_size > 1) {
            shift -= 2;
            if (shift <= -swap_size) {
                shift = swap_size - 1;
            }
        }
    }

    return nread;
}

long spReadID3UnsynchronizedULong32(unsigned long *value, int unsync_flag,
                                    unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nread;
    
    if ((nread = spReadID3UnsynchronizedBuffer((unsigned char *)value, 4, 0, 4, 4, SP_AIFF_NEED_SWAP,
                                               unsync_flag, last_value, unsync_count, fp)) <= 0) {
        spDebug(10, "spReadID3UnsynchronizedULong32", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }

    return nread;
}

long spWriteID3UnsynchronizedBuffer(unsigned char *buf, long buf_len, long block_len, long stride,
                                    int swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long k;
    long ret;
    long nwrite;
    unsigned char prev_value;
    unsigned char c, zc;
    long shift;
    long swap_size = 0;

    zc = 0;
    prev_value = 0;
    if (last_value != NULL) prev_value = *last_value;

    if (swap_flag) {
        swap_size = block_len;
    }
    
    if (swap_size > 1) {
        shift = swap_size - 1;
    } else {
        shift = 0;
    }

    for (k = 0, nwrite = 0;;) {
        if (!SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == 0) k += (stride - block_len);
        }

        c = buf[k + shift];

        if (unsync_flag) {
            if (prev_value == 0xff && ((c & 0xe0) == 0xe0 || c == 0)) { /* 0xe0 = 11100000 */
                zc = 0;
                if ((ret = (long)fwrite(&zc, 1, 1, fp)) != 1) {
                    spDebug(10, "spWriteID3UnsynchronizedBuffer", "fwrite failed: %ld\n", ret);
                    return ret;
                }
                if (last_value != NULL) *last_value = zc;
                if (unsync_count != NULL) {
                    *unsync_count += 1;
                }
                spDebug(10, "spWriteID3UnsynchronizedBuffer", "unsync: %ld\n", nwrite);
                ++nwrite;
            }
        }

        if ((ret = (long)fwrite(&c, 1, 1, fp)) != 1) {
            spDebug(10, "spWriteID3UnsynchronizedBuffer", "fwrite failed: %ld\n", ret);
            return ret;
        }
        if (last_value != NULL) *last_value = c;
        ++nwrite;

        ++k;

        if (SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == block_len) k += (stride - block_len);
        }

        if (k >= buf_len) {
            break;
        }
        
        if (swap_size > 1) {
            shift -= 2;
            if (shift <= -swap_size) {
                shift = swap_size - 1;
            }
        }
        
        prev_value = c;
    }

    return nwrite;
}

long spWriteID3UnsynchronizedULong32(unsigned long value, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nwrite;

    if ((nwrite = spWriteID3UnsynchronizedBuffer((unsigned char *)&value, 4, 4, 4, SP_AIFF_NEED_SWAP, unsync_flag,
                                                 last_value, unsync_count, fp)) <= 0) {
        spDebug(10, "spWriteID3UnsynchronizedULong32", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }

    return nwrite;
}

spID3FrameSize spGetID3FrameContentSizeForChunkFileSpec(spID3Frame *frame)
{
    return frame->header.size;
}

spID3FrameSize spGetID3FrameContentSize(spID3Frame *frame)
{
    if (frame == NULL) return 0;
    
    return spGetID3FrameContentSizeForChunkFileSpec(frame);
}

spBool spSetID3FrameContentSizeForChunkFileSpec(spID3Frame *frame, spID3FrameSize size)
{
    frame->header.size = (unsigned long)size;

    return SP_TRUE;
}

spID3FrameSize spGetID3FrameHeaderExtraSize(spID3Frame *frame)
{
    spID3Header *header;
    spID3FrameSize size = 0;
    
    header = spGetID3HeaderFromFrame(frame);
    
    if (header->version == 3) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
            size += 4;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
            size += 1;
        }
    } else if (header->version == 4) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
            size += 4;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_DATA_LENGTH_MASK) {
            size += 4;
        }
    }
        
    return size;
}

spID3FrameSize spGetID3FrameHeaderSize(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);
    
    if (header->version == 2) {
        return 6;
    } else {
        return 10 + spGetID3FrameHeaderExtraSize(frame);
    }
}

spID3FrameSize spGetID3FrameSize(spID3Frame *frame, char *string)
{
    if (string != NULL) {
        memcpy(string, frame->header.type, 4);
        string[4] = NUL;
    }

    return frame->header.size + spGetID3FrameHeaderSize(frame);
}

spID3FrameSize spGetID3HeaderSize(spID3Header *header)
{
    return spGetID3FrameSize((spID3Frame *)header, NULL);
}

spID3FrameSize spReadID3FrameHeader(spID3Frame *parent, spID3FrameHeader *frame_header, int swap, FILE *fp)
{
    long len;
    spID3FrameSize nread;
    spID3Header *header;
    unsigned char last_value = 0;

    header = spGetID3HeaderFromFrame(parent);

    memset(frame_header, 0, sizeof(spID3FrameHeader));

    if (header->version == 2) {
        len = 3;
        frame_header->type[3] = ' ';
    } else {
        len = 4;
    }
    
    if (fread(frame_header->type, 1, len, fp) != len) {
        spDebug(10, "spReadID3FrameHeader", "Can't read frame type.\n");
        return 0;
    }
    nread = len;
    spDebug(10, "spReadID3FrameHeader", "len = %d, type = %c%c%c%c\n",
            len, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);
    
    if (header->version == 2) {
        if ((len = spReadID3UnsynchronizedBuffer((unsigned char *)&frame_header->size, 3, 0, 3, 4, SP_AIFF_NEED_SWAP,
                                                 header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                 &last_value, &header->unsync_count, fp)) <= 0) {
            spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        nread += 3;
    } else {
        if (header->version == 3) {
            if ((len = spReadID3UnsynchronizedULong32(&frame_header->size,
                                                      header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                      &last_value, &header->unsync_count, fp)) <= 0) {
                spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedULong32 failed\n");
                return 0;
            }
        } else {
            if (spReadID3SyncSafeLong32(&frame_header->size, fp) != 4) {
                spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
                return 0;
            }
        }
        nread += 4;
        spDebug(10, "spReadID3FrameHeader", "size = %ld\n", frame_header->size);

        if (fread(&frame_header->flag1, 1, 1, fp) != 1
            || fread(&frame_header->flag2, 1, 1, fp) != 1) {
            return 0;
        }
        nread += 2;
        spDebug(10, "spReadID3FrameHeader", "flag1 = %lx, flag2 = %lx\n", frame_header->flag1, frame_header->flag2);
        
        if (header->version == 3) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
                if ((len = spReadID3UnsynchronizedULong32(&frame_header->extra.v2_3.decompressed_size,
                                                          header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                          &last_value, &header->unsync_count, fp)) <= 0) {
                    spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedULong32 failed\n");
                    return 0;
                }
                nread += 4;
                spDebug(10, "spReadID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_3.decompressed_size);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
                if (fread(&frame_header->extra.v2_3.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_3.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
                if (fread(&frame_header->extra.v2_3.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_3.group_id);
            }
        } else if (header->version == 4) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
                if (fread(&frame_header->extra.v2_4.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_4.group_id);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
                if (fread(&frame_header->extra.v2_4.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_4.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
                if (spReadID3SyncSafeLong32(&frame_header->extra.v2_4.data_length_indicator, fp) != 4) {
                    spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
                    return 0;
                }
                nread += 4;
                spDebug(10, "spReadID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_4.data_length_indicator);
            }
        }
    }
    spDebug(10, "spReadID3FrameHeader", "size = %ld\n", frame_header->size);
    
    return nread;
}

spID3FrameSize spWriteID3FrameHeader(spID3Frame *frame, spID3FrameHeader *frame_header, int swap, FILE *fp)
{
    long len;
    spID3FrameSize nwrite;
    spID3Header *header;
    unsigned long size;
    unsigned char last_value = 0;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version == 2) {
        len = 3;
    } else {
        len = 4;
    }
    
    spDebug(10, "spWriteID3FrameHeader", "len = %d, type = %c%c%c%c\n",
            len, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);
    if (fwrite(frame_header->type, 1, len, fp) != len) {
        spDebug(10, "spWriteID3FrameHeader", "Can't write frame type.\n");
        return 0;
    }
    nwrite = len;

    if (spIsID3NullFrameHeader(frame_header)) {
        size = 0;
    } else {
        size = frame_header->size;
    }
    
    spDebug(10, "spWriteID3FrameHeader", "size = %ld\n", size);
    
    if (header->version == 2) {
        if ((len = spWriteID3UnsynchronizedBuffer((unsigned char *)&size, 3, 3, 4, SP_AIFF_NEED_SWAP,
                                                  header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                  &last_value, &header->unsync_count, fp)) <= 0) {
            spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        nwrite += 3;
    } else {
        if (header->version == 3) {
            if ((len = spWriteID3UnsynchronizedULong32(size,
                                                       header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                       &last_value, &header->unsync_count, fp)) <= 0) {
                spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedULong32 failed\n");
                return 0;
            }
        } else {
            if (spWriteID3SyncSafeLong32(size, fp) != 4) {
                spDebug(10, "spWriteID3Header", "Can't write size: %ld\n", (long)nwrite);
                return 0;
            }
        }
        nwrite += 4;

        spDebug(10, "spWriteID3FrameHeader", "flag1 = %lx, flag2 = %lx\n", frame_header->flag1, frame_header->flag2);
        if (fwrite(&frame_header->flag1, 1, 1, fp) != 1
            || fwrite(&frame_header->flag2, 1, 1, fp) != 1) {
            return 0;
        }
        nwrite += 2;
        
        if (header->version == 3) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
                if ((len = spWriteID3UnsynchronizedULong32(frame_header->extra.v2_3.decompressed_size,
                                                           header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                           &last_value, &header->unsync_count, fp)) <= 0) {
                    spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedULong32 failed\n");
                    return 0;
                }
                nwrite += 4;
                spDebug(10, "spWriteID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_3.decompressed_size);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
                if (fwrite(&frame_header->extra.v2_3.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_3.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
                if (fwrite(&frame_header->extra.v2_3.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_3.group_id);
            }
        } else if (header->version == 4) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
                if (fwrite(&frame_header->extra.v2_4.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_4.group_id);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
                if (fwrite(&frame_header->extra.v2_4.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_4.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
                if (spWriteID3SyncSafeLong32(frame_header->extra.v2_4.data_length_indicator, fp) != 4) {
                    spDebug(10, "spWriteID3Header", "Can't write size: %ld\n", (long)nwrite);
                    return 0;
                }
                nwrite += 4;
                spDebug(10, "spWriteID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_4.data_length_indicator);
            }
        }
    }
    
    spDebug(10, "spWriteID3FrameHeader", "done: size = %ld\n", frame_header->size);
    
    return nwrite;
}

spID3FrameSize spSkipID3Frame(spID3Frame *parent, spID3FrameHeader *frame_header, spID3FrameSize remain_size, int swap, FILE *fp)
{
    spID3FrameSize skiplen;
    spID3Header *header;

    header = spGetID3HeaderFromFrame(parent);
    
    spDebug(10, "spSkipID3Frame", "remain_size = %ld, frame_header->size = %ld, frame_header->type = %c%c%c%c\n",
            (long)remain_size, frame_header->size, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);

    if (spIsID3NullFrameHeader(frame_header)) {
        skiplen = remain_size;
        spDebug(10, "spSkipID3Frame", "**** padding found ****, skiplen = %ld\n", (long)skiplen);
        if (spSeekFile(fp, skiplen, SEEK_CUR) != 0) {
            return 0;
        }
        frame_header->size = (unsigned long)skiplen;
    } else {
    
        if (frame_header->size <= 0) {
            skiplen = 0;
        } else {
            if (header->version == 4 || !(header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK)) {
                skiplen = frame_header->size;
                if (spSeekFile(fp, skiplen, SEEK_CUR) != 0) {
                    return 0;
                }
            } else {
                unsigned char c, pc;
                long len;

                pc = 0;
                for (skiplen = 0;;) {
                    if ((len = spReadID3UnsynchronizedBuffer(&c, 1, 0, 1, 1, SP_FALSE, SP_TRUE, &pc, &header->unsync_count, fp)) <= 0) {
                        return 0;
                    }

                    if (++skiplen >= frame_header->size) {
                        break;
                    }
                    pc = c;
                }
            }
        }
    }

    return skiplen;
}

spID3FrameSize spWriteID3ContentRest(spID3Frame *frame, long depth, spID3FrameSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    spID3FrameSize len;
    spID3FrameSize nwrite;
    spID3FrameSize nremain;
    spID3FrameSize total_nwrite;
    static unsigned char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};

    spDebug(80, "spWriteID3ContentRest", "frame->header.size = %ld, frame->header.type = %c%c%c%c\n",
            frame->header.size, frame->header.type[0], frame->header.type[1], frame->header.type[2], frame->header.type[3]);
    
    spDebug(80, "spWriteID3ContentRest", "depth = %ld, remain_size %ld\n", depth, (long)remain_size);

    total_nwrite = 0;
    nremain = remain_size;
    
    spDebug(80, "spWriteID3ContentRest", "nremain = %ld\n", (long)nremain);
    
    while (nremain > 0) {
        len = MIN(8, nremain);
        spDebug(100, "spWriteID3ContentRest", "len = %ld, nremain = %ld\n", (long)len, (long)nremain);
        if ((nwrite = fwrite(buf, 1, (long)len, fp)) != len) {
            spDebug(10, "spWriteID3ContentRest", "fwrite zeros failed: nwrite = %ld\n", (long)nwrite);
            break;
        }
        spDebug(100, "spWriteID3ContentRest", "fwrite zeros done: nwrite = %ld\n", (long)nwrite);

        nremain -= nwrite;
        total_nwrite += nwrite;
    }

    spDebug(80, "spWriteID3ContentRest", "done: total_nwrite = %ld\n", (long)total_nwrite);
    
    return total_nwrite;
}

long spReadID3String(unsigned char encoding, unsigned char *buf, long need_len, long read_len, spBool *bom_flag,
                     spBool *swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long len;
    long nread;
    long uc;

    uc = 0;
    if ((nread = spReadID3UnsynchronizedBuffer(buf, need_len, read_len, 1, 1, SP_FALSE, unsync_flag, last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3String", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    len = nread - uc;
    if (unsync_count != NULL) {
        *unsync_count += uc;
    }
    
    if (bom_flag != NULL) *bom_flag = SP_FALSE;
    
    if (encoding == 0/* ISO-88159-1 */ || encoding == 3/* UTF-8 */) {
        if (swap_flag != NULL) *swap_flag = SP_FALSE;
    } else {
        int swap = 0;
        unsigned short *sbuf = (unsigned short *)buf;

        if (encoding != 2) {        /* with BOM */
            if (sbuf[0] == 0xfffe) {
                swap = 1;
            }
            
            if (sbuf[0] == 0xfffe || sbuf[0] == 0xfeff) {
                long k;

                /*nread -= 2;*/
                len -= 2;
                
                for (k = 0; k < len / 2; k++) {
                    sbuf[k] = sbuf[k + 1];
                }
                sbuf[k] = 0;

                if (bom_flag != NULL) *bom_flag = SP_TRUE;
            } else {
                swap = SP_AIFF_NEED_SWAP; /* no BOM */
            }
        } else {        /* 2: without BOM */
            swap = SP_AIFF_NEED_SWAP;
        }

        if (swap) {
            spSwapShort((short *)sbuf, len / 2);
        }
        
        if (swap_flag != NULL) *swap_flag = swap;
    }

    return nread;
}

long spWriteID3String(unsigned char encoding, unsigned char *buf, long buf_len, spBool bom_flag,
                      spBool swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nwrite;
    long uc;
    long bom_size = 0;
    
    spDebug(80, "spWriteID3String", "encoding = %d, buf_len = %ld, bom_flag = %d, swap_flag = %d\n",
            encoding, buf_len, bom_flag, swap_flag);
    
    if (encoding == 0/* ISO-88159-1 */ || encoding == 3/* UTF-8 */) {
        /* do nothing */
    } else {
        if (bom_flag && encoding != 2) {        /* with BOM */
            short bom = 0xfeff;

            uc = 0;
            if ((nwrite = spWriteID3UnsynchronizedBuffer((unsigned char *)&bom, 2, 2, 2, swap_flag, unsync_flag, last_value, &uc, fp)) <= 0) {
                spDebug(10, "spWriteID3String", "Can't write BOM of UTF string.\n");
                return nwrite;
            }
            if (unsync_count != NULL) {
                *unsync_count += uc;
            }
            bom_size = 2;
        }
    }

    uc = 0;
    if ((nwrite = spWriteID3UnsynchronizedBuffer(buf, buf_len, 2, 2, swap_flag, unsync_flag, last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3String", "spWriteID3UnsynchronizedBuffer failed\n");
    } else {
        if (unsync_count != NULL) {
            *unsync_count += uc;
        }
    }
    
    spDebug(80, "spWriteID3String", "done: nwrite = %ld\n", nwrite);
    
    if (nwrite > 0) {
        return nwrite + bom_size;
    } else {
        return nwrite;
    }
}

spID3FrameSize spReadID3DataFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    spDebug(80, "spReadID3DataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    dframe->data = xalloc(dframe->header.size + 1, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = dframe->header.size;
    } else {
        need_len = dframe->header.size;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3UnsynchronizedBuffer(dframe->data, need_len, read_len, 1, 1, SP_FALSE, SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3DataFrame", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    dframe->data[len - uc] = 0;        /* just in case */
    header->unsync_count += uc;
    nread = len;
    
    spDebug(80, "spReadID3DataFrame", "done: nread = %ld / %ld\n", (long)nread, dframe->header.size);
    
    return nread;
}

spID3FrameSize spWriteID3DataFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long uc;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    spDebug(80, "spWriteID3DataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    uc = 0;
    if ((len = spWriteID3UnsynchronizedBuffer(dframe->data, dframe->header.size, 1, 1, 0, 1, 0, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3DataFrame", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)dframe);
    nwrite = len;
    
    spDebug(80, "spWriteID3DataFrame", "done: nwrite = %ld / %ld\n", (long)nwrite, dframe->header.size);
    
    return nwrite;
}

spBool spFreeID3DataFrame(void *frame)
{
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    if (dframe->data != NULL) {
        xfree(dframe->data);
        dframe->data = NULL;
    }
    
    return SP_TRUE;
}

spID3FrameSize spCopyID3DataFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    spID3DataFrame *dest_dframe = (spID3DataFrame *)dest_frame;
    spID3DataFrame *src_dframe = (spID3DataFrame *)src_frame;

    spDebug(80, "spCopyID3DataFrame", "in\n");

    dest_dframe->data = xalloc(src_dframe->header.size + 1, unsigned char);
    memcpy(dest_dframe->data, src_dframe->data, src_dframe->header.size);
    dest_dframe->data[src_dframe->header.size] = 0;        /* just in case */
    
    return src_dframe->header.size;
}

spID3FrameSize spReadID3TextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    spDebug(80, "spReadID3TextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    if (fread(&tframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "spReadID3TextFrame", "encoding = %d\n", tframe->encoding);

    tframe->string = xalloc(tframe->header.size - 1 + 2, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = tframe->header.size - 1;
    } else {
        need_len = tframe->header.size - 1;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3String(tframe->encoding, tframe->string, need_len, read_len, &tframe->bom_flag, &tframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)tframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3TextFrame", "Can't read string.\n");
        return 0;
    }
    tframe->string[len - uc] = 0;
    tframe->string[len - uc + 1] = 0;
    header->unsync_count += uc;
    nread += len;
    spDebug(80, "spReadID3TextFrame", "string = `%s'\n", tframe->string);
    
    spDebug(80, "spReadID3TextFrame", "done: nread = %ld / %ld\n", (long)nread, tframe->header.size);
    
    return nread;
}

spID3FrameSize spWriteID3TextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long write_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    spDebug(80, "spWriteID3TextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "spWriteID3TextFrame", "encoding = %d\n", tframe->encoding);
    if (fwrite(&tframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    write_len = tframe->header.size - 1;
    if (tframe->bom_flag) {
        write_len -= 2;
    }

    spDebug(80, "spWriteID3TextFrame", "write_len = %ld, string = `%s'\n", write_len, tframe->string);
    uc = 0;
    if ((len = spWriteID3String(tframe->encoding, tframe->string, write_len, tframe->bom_flag, tframe->swap_flag,
                                spIsID3FrameUnsynchronized((spID3Frame *)tframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3TextFrame", "Can't write string.\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)tframe);
    nwrite += len;
    
    spDebug(80, "spWriteID3TextFrame", "done: nwrite = %ld / %ld, unsync_count = %ld\n",
            (long)nwrite, tframe->header.size, header->unsync_count);
    
    return nwrite;
}

static spID3FrameSize readID3LangTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp, spBool lang_flag)
{
    long len;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3LangTextFrame *ltframe = (spID3LangTextFrame *)frame;

    spDebug(80, "readID3LangTextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    ltframe->string = NULL;
    ltframe->actual_string = NULL;
    
    if (fread(&ltframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "readID3LangTextFrame", "encoding = %d\n", ltframe->encoding);

    if (lang_flag) {
        if (fread(ltframe->lang, 1, 3, fp) != 3) {
            return 0;
        }
        nread += 3;
        spDebug(80, "readID3LangTextFrame", "lang = %c%c%c\n", ltframe->lang[0], ltframe->lang[1], ltframe->lang[2]);
    }

    len = ltframe->header.size - (long)nread;
    ltframe->string = xalloc(len + 2, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = len;
    } else {
        need_len = len;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3String(ltframe->encoding, ltframe->string, need_len, read_len, &ltframe->bom_flag, &ltframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)ltframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "readID3LangTextFrame", "Can't read string.\n");
        return 0;
    }
    ltframe->string[len - uc] = 0;
    ltframe->string[len - uc + 1] = 0;
    header->unsync_count += uc;
    nread += len;
    spDebug(80, "readID3LangTextFrame", "string = `%s'\n", ltframe->string);

    if (strneq(ltframe->header.type, "COM ", 4)
        || strneq(ltframe->header.type, "COMM", 4)
        || strneq(ltframe->header.type, "ULT ", 4)
        || strneq(ltframe->header.type, "USLT", 4)
        || strneq(ltframe->header.type, "TXX ", 4)
        || strneq(ltframe->header.type, "TXXX", 4)
        || strneq(ltframe->header.type, "WXX ", 4)
        || strneq(ltframe->header.type, "WXXX", 4)
        ) {
        int i;

        for (i = (int)strlen((char *)ltframe->string) + 1; i < len - uc; i++) {
            if (ltframe->string[i] != 0) {
                ltframe->actual_string = &ltframe->string[i];
                break;
            }
        }
    }
    
    spDebug(80, "readID3LangTextFrame", "done: nread = %ld / %ld\n", (long)nread, ltframe->header.size);
    
    return nread;
}

spID3FrameSize spReadID3DescTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    return readID3LangTextFrame(parent, depth, frame, swap, fp, SP_FALSE);
}

spID3FrameSize spReadID3LangTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    return readID3LangTextFrame(parent, depth, frame, swap, fp, SP_TRUE);
}

static spID3FrameSize writeID3LangTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp, spBool lang_flag)
{
    long len;
    long write_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3LangTextFrame *ltframe = (spID3LangTextFrame *)frame;

    spDebug(80, "writeID3LangTextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "writeID3LangTextFrame", "encoding = %d\n", ltframe->encoding);
    if (fwrite(&ltframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    if (lang_flag) {
        spDebug(80, "writeID3LangTextFrame", "lang = %c%c%c\n", ltframe->lang[0], ltframe->lang[1], ltframe->lang[2]);
        if (fwrite(ltframe->lang, 1, 3, fp) != 3) {
            return 0;
        }
        nwrite += 3;
    }

    write_len = ltframe->header.size - (long)nwrite;
    if (ltframe->bom_flag) {
        write_len -= 2;
    }

    spDebug(80, "writeID3LangTextFrame", "write_len = %ld, string = `%s'\n", write_len, ltframe->string);
    uc = 0;
    if ((len = spWriteID3String(ltframe->encoding, ltframe->string, write_len, ltframe->bom_flag, ltframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)ltframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "writeID3LangTextFrame", "Can't write string.\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)ltframe);
    nwrite += len;

    spDebug(80, "writeID3LangTextFrame", "done: nwrite = %ld / %ld, unsync_count = %ld\n",
            (long)nwrite, ltframe->header.size, header->unsync_count);
    
    return nwrite;
}

spID3FrameSize spWriteID3DescTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    return writeID3LangTextFrame(frame, depth, rewrite_flag, swap, fp, SP_FALSE);
}

spID3FrameSize spWriteID3LangTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    return writeID3LangTextFrame(frame, depth, rewrite_flag, swap, fp, SP_TRUE);
}

spBool spFreeID3TextFrame(void *frame)
{
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    if (tframe->string != NULL) {
        xfree(tframe->string);
        tframe->string = NULL;
    }
    
    return SP_TRUE;
}

static long copyID3String(unsigned char **dest_pstring, long *dest_len, unsigned char **src_pstring, long src_len)
{
    long ncopy = 0;
    unsigned char *dest_string = NULL;
    
    if (*src_pstring != NULL) {
        *dest_len = src_len;
        dest_string = xalloc(src_len + 2, unsigned char);
        memcpy(dest_string, *src_pstring, src_len);
        dest_string[src_len] = NUL;
        dest_string[src_len + 1] = NUL;        /* just in case */
        ncopy += src_len;
    } else {
        *dest_len = 0;
    }

    *dest_pstring = dest_string;

    ncopy += sizeof(long) + sizeof(unsigned char *);

    return ncopy;
}

spID3FrameSize spCopyID3TextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    long dest_len = 0;
    spID3FrameSize ncopy;
    spID3TextFrame *dest_tframe = (spID3TextFrame *)dest_frame;
    spID3TextFrame *src_tframe = (spID3TextFrame *)src_frame;

    dest_tframe->encoding = src_tframe->encoding;
    dest_tframe->bom_flag = src_tframe->bom_flag;
    dest_tframe->swap_flag = src_tframe->swap_flag;
    ncopy = sizeof(dest_tframe->encoding) + sizeof(dest_tframe->bom_flag) + sizeof(dest_tframe->swap_flag);

    ncopy += copyID3String(&dest_tframe->string, &dest_len, &src_tframe->string, src_tframe->header.size - 1);
    
    return ncopy;
}

static spID3FrameSize copyID3LangTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth, spBool lang_flag)
{
    long len;
    spID3FrameSize ncopy;
    spID3LangTextFrame *dest_ltframe = (spID3LangTextFrame *)dest_frame;
    spID3LangTextFrame *src_ltframe = (spID3LangTextFrame *)src_frame;

    spDebug(80, "copyID3LangTextFrame", "in\n");

    dest_ltframe->string = NULL;
    dest_ltframe->actual_string = NULL;

    dest_ltframe->encoding = src_ltframe->encoding;
    len = src_ltframe->header.size - 1;
    ncopy = sizeof(dest_ltframe->encoding);

    dest_ltframe->bom_flag = src_ltframe->bom_flag;
    dest_ltframe->swap_flag = src_ltframe->swap_flag;
    ncopy += sizeof(dest_ltframe->bom_flag) + sizeof(dest_ltframe->swap_flag);
    
    if (lang_flag) {
        memcpy(dest_ltframe->lang, src_ltframe->lang, 3);
        ncopy += 3;
        len -= 3;
    }

    dest_ltframe->string = xalloc(len + 2, unsigned char);
    memcpy(dest_ltframe->string, src_ltframe->string, len);
    dest_ltframe->string[len] = 0;
    dest_ltframe->string[len + 1] = 0;
    ncopy += len;

    if (strneq(src_ltframe->header.type, "COM ", 4)
        || strneq(src_ltframe->header.type, "COMM", 4)
        || strneq(src_ltframe->header.type, "ULT ", 4)
        || strneq(src_ltframe->header.type, "USLT", 4)
        || strneq(src_ltframe->header.type, "TXX ", 4)
        || strneq(src_ltframe->header.type, "TXXX", 4)
        || strneq(src_ltframe->header.type, "WXX ", 4)
        || strneq(src_ltframe->header.type, "WXXX", 4)
        ) {
        int i;

        for (i = (int)strlen((char *)dest_ltframe->string) + 1; i < len; i++) {
            if (dest_ltframe->string[i] != 0) {
                dest_ltframe->actual_string = &dest_ltframe->string[i];
                break;
            }
        }
    }
    
    return ncopy;
}

spID3FrameSize spCopyID3DescTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    return copyID3LangTextFrame(dest_frame, src_frame, depth, SP_FALSE);
}

spID3FrameSize spCopyID3LangTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    return copyID3LangTextFrame(dest_frame, src_frame, depth, SP_TRUE);
}

long spReadUnknownLengthString(unsigned char **pstring, long alloc_incr,
                               long remain_size, long read_limit,
                               int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    unsigned char c;
    long len;
    long nread;
    long nchar;
    long alloc_len;
    unsigned char *string;

    alloc_incr = MAX(alloc_incr, 16);
    alloc_len = alloc_incr;
    
    *pstring = xalloc(alloc_len, unsigned char);
    string = *pstring;

    for (nchar = 0, nread = 0;;) {
        if ((len = spReadID3UnsynchronizedBuffer(&c, 1, 0, 1, 1, SP_FALSE, unsync_flag, last_value, unsync_count, fp)) <= 0) {
            return 0;
        }
        nread += len;

        if (c == NUL) {
            break;
        }

        if (remain_size > 0 && ++nchar >= remain_size) {
            break;
        }
        if (read_limit > 0 && nread >= read_limit) {
            break;
        }

        if (nchar + 2 >= alloc_len) {
            alloc_len += alloc_incr;
            *pstring = xrealloc(*pstring, alloc_len, unsigned char);
            string = *pstring;
        }
        string[nchar - 1] = c;
    }
    
    string[nchar] = NUL;
    string[nchar + 1] = NUL;        /* just in case */

    return nread;
}

spID3FrameSize spReadID3MimeDataFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len;
    long uc;
    long remain_size, read_limit;
    unsigned char last_value = 0;
    spBool unsynchronized;
    long nread;
    spID3Header *header;
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    spDebug(80, "spReadID3MimeDataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    unsynchronized = spIsID3FrameUnsynchronized((spID3Frame *)mdframe);
    
    memset(mdframe->format, 0, 3);
    mdframe->mime_type_len = 0;
    mdframe->mime_type = NULL;
    mdframe->filename_len = 0;
    mdframe->filename = NULL;
    mdframe->data_type = 0;
    mdframe->description_len = 0;
    mdframe->description = NULL;
    mdframe->data_size = 0;
    mdframe->data = NULL;
    
    if (fread(&mdframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "spReadID3MimeDataFrame", "encoding = %d\n", mdframe->encoding);

    if (strneq(mdframe->header.type, "PIC ", 4)) {
        if (fread(mdframe->format, 1, 3, fp) != 3) {
            return 0;
        }
        nread += 3;
        spDebug(80, "spReadID3MimeDataFrame", "format = %c%c%c\n", mdframe->format[0], mdframe->format[1], mdframe->format[2]);
    } else {
        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->header.size - nread;
        } else {
            remain_size = mdframe->header.size - nread;
            read_limit = 0;
        }

        uc = 0;
        if ((len = spReadUnknownLengthString(&mdframe->mime_type, 16, remain_size, read_limit,
                                             unsynchronized, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        mdframe->mime_type_len = len - uc;
        nread += len;
        spDebug(80, "spReadID3MimeDataFrame", "len = %d, mime_type = %s\n", len, mdframe->mime_type);
    }

    if (strneq(mdframe->header.type, "GEO", 3)) {
        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->header.size - nread;
        } else {
            remain_size = mdframe->header.size - nread;
            read_limit = 0;
        }

        uc = 0;
        if ((len = spReadUnknownLengthString(&mdframe->filename, 16, remain_size, read_limit,
                                             unsynchronized, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        mdframe->filename_len = len - uc;
        nread += len;
        spDebug(80, "spReadID3MimeDataFrame", "len = %d, filename = %s\n", len, mdframe->filename);
    } else {
        if (fread(&mdframe->data_type, 1, 1, fp) != 1) {
            return 0;
        }
        nread += 1;
        spDebug(80, "spReadID3MimeDataFrame", "data_type = %d\n", mdframe->data_type);
    }

    if (header->version == 4) {
        remain_size = 0;
        read_limit = mdframe->header.size - nread;
    } else {
        remain_size = mdframe->header.size - nread;
        read_limit = 0;
    }

    uc = 0;
    if ((len = spReadUnknownLengthString(&mdframe->description, 16, remain_size, read_limit,
                                         unsynchronized, &last_value, &uc, fp)) <= 0) {
        spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
        return 0;
    }
    header->unsync_count += uc;
    mdframe->description_len = len - uc;
    nread += len;
    spDebug(80, "spReadID3MimeDataFrame", "len = %d, description = %s\n", len, mdframe->description);

    mdframe->data_size = mdframe->header.size - nread;
    mdframe->data = xalloc(mdframe->data_size, unsigned char);

    if (unsynchronized == SP_TRUE) {
        spDebug(10, "spReadID3MimeDataFrame", "**** unsynchrozisation ****\n");

        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->data_size;
        } else {
            remain_size = mdframe->data_size;
            read_limit = 0;
        }
    
        uc = 0;
        if ((len = spReadID3UnsynchronizedBuffer(mdframe->data, remain_size, read_limit, 1, 1, SP_FALSE,
                                                 SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(10, "spReadID3MimeDataFrame", "spReadID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        if (header->version == 4) {
            mdframe->data_size -= uc;
        }
        header->unsync_count += uc;
        nread += len;
    } else {
        spDebug(10, "spReadID3MimeDataFrame", "---- no unsynchrozisation ----\n");
        if (fread(mdframe->data, 1, mdframe->data_size, fp) != mdframe->data_size) {
            return 0;
        }
        nread += mdframe->data_size;
    }
    
    spDebug(80, "spReadID3MimeDataFrame", "data_size = %ld\n", mdframe->data_size);

    spDebug(80, "spReadID3MimeDataFrame", "done: nread = %ld / %ld\n", (long)nread, mdframe->header.size);
    
    return nread;
}

long spWriteUnknownLengthString(unsigned char *string, long len, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    return spWriteID3UnsynchronizedBuffer(string, len, 1, 1, SP_FALSE, unsync_flag, last_value, unsync_count, fp);
}

spID3FrameSize spWriteID3MimeDataFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    spDebug(80, "spWriteID3MimeDataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "spWriteID3MimeDataFrame", "encoding = %d\n", mdframe->encoding);
    if (fwrite(&mdframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    if (strneq(mdframe->header.type, "PIC ", 4)) {
        spDebug(80, "spWriteID3MimeDataFrame", "format = %c%c%c\n", mdframe->format[0], mdframe->format[1], mdframe->format[2]);
        if (fwrite(mdframe->format, 1, 3, fp) != 3) {
            return 0;
        }
        nwrite += 3;
    } else {
        spDebug(80, "spWriteID3MimeDataFrame", "len = %d, mime_type = %s\n", mdframe->mime_type_len, mdframe->mime_type);
        uc = 0;
        if ((len = spWriteUnknownLengthString(mdframe->mime_type, mdframe->mime_type_len, 
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        nwrite += len;
    }

    if (strneq(mdframe->header.type, "GEO", 3)) {
        spDebug(80, "spWriteID3MimeDataFrame", "len = %d, filename = %s\n", mdframe->filename_len, mdframe->filename);
        uc = 0;
        if ((len = spWriteUnknownLengthString(mdframe->filename, mdframe->filename_len,
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        nwrite += len;
    } else {
        spDebug(80, "spWriteID3MimeDataFrame", "data_type = %d\n", mdframe->data_type);
        if (fwrite(&mdframe->data_type, 1, 1, fp) != 1) {
            return 0;
        }
        nwrite += 1;
    }

    spDebug(80, "spWriteID3MimeDataFrame", "len = %d, description = %s\n", mdframe->description_len, mdframe->description);
    uc = 0;
    if ((len = spWriteUnknownLengthString(mdframe->description, mdframe->description_len,
                                          SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
        return 0;
    }
    header->unsync_count += uc;
    nwrite += len;

    spDebug(80, "spWriteID3MimeDataFrame", "data_size = %ld\n", mdframe->data_size);
    uc = 0;
    if ((len = spWriteID3UnsynchronizedBuffer(mdframe->data, mdframe->data_size, 1, 1, SP_FALSE,
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3MimeDataFrame", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    header->unsync_count += uc;
    nwrite += len;
    
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)mdframe);
    
    spDebug(80, "spWriteID3MimeDataFrame", "done: nwrite = %ld / %ld\n", (long)nwrite, mdframe->header.size);
    
    return nwrite;
}

spID3FrameSize spCopyID3MimeDataFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    long ncopy;
    spID3MimeDataFrame *dest_mdframe = (spID3MimeDataFrame *)dest_frame;
    spID3MimeDataFrame *src_mdframe = (spID3MimeDataFrame *)src_frame;

    spDebug(80, "spCopyID3MimeDataFrame", "in\n");

    dest_mdframe->encoding = src_mdframe->encoding;
    ncopy = 1;

    memcpy(dest_mdframe->format, src_mdframe->format, 3);
    ncopy += 3;
    
    ncopy += copyID3String(&dest_mdframe->mime_type, &dest_mdframe->mime_type_len,
                           &src_mdframe->mime_type, src_mdframe->mime_type_len);

    ncopy += copyID3String(&dest_mdframe->filename, &dest_mdframe->filename_len,
                           &src_mdframe->filename, src_mdframe->filename_len);

    dest_mdframe->data_type = src_mdframe->data_type;
    ncopy += 1;

    ncopy += copyID3String(&dest_mdframe->description, &dest_mdframe->description_len,
                           &src_mdframe->description, src_mdframe->description_len);

    dest_mdframe->data_size = src_mdframe->data_size;
    dest_mdframe->data = xalloc(dest_mdframe->data_size, unsigned char);
    memcpy(dest_mdframe->data, src_mdframe->data, dest_mdframe->data_size);
    ncopy += dest_mdframe->data_size;
    
    return ncopy;
}

spBool spFreeID3MimeDataFrame(void *frame)
{
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    if (mdframe->mime_type != NULL) {
        xfree(mdframe->mime_type);
        mdframe->mime_type = NULL;
    }
    
    if (mdframe->filename != NULL) {
        xfree(mdframe->filename);
        mdframe->filename = NULL;
    }
    
    if (mdframe->description != NULL) {
        xfree(mdframe->description);
        mdframe->description = NULL;
    }
    
    if (mdframe->data != NULL) {
        xfree(mdframe->data);
        mdframe->data = NULL;
    }
    
    return SP_TRUE;
}

static spChunkInfoTable sp_id3_frame_info_list[] = {
    {"", "ID3", SP_FALSE, SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
    {"TXX ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"TXXX", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"WXX ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"WXXX", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"T", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_TEXT_MASK, 0, sizeof(spID3TextFrame), 0, spReadID3TextFrame, spWriteID3TextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3TextFrame},
    {"W", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
    {"UFI", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
    {"COM ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"COMM", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"USER", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"ULT ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"USLT", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"PIC ", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"APIC", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"GEO", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"PRIV", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
};

spChunkFileSpec sp_id3_file_spec = {
    sizeof(struct _spID3FrameHeader),
    sizeof(struct _spID3Frame),
    sp_id3_frame_info_list,
    0,

    NULL,
    NULL,
    NULL,
    (spChunkGetSizeFunc)spGetID3FrameHeaderSize,
    (spChunkGetSizeFunc)spGetID3FrameContentSizeForChunkFileSpec,
    (spChunkSetSizeFunc)spSetID3FrameContentSizeForChunkFileSpec,
    NULL,
    NULL,
    (spChunkReadHeaderFunc)spReadID3FrameHeader,
    NULL,
    (spChunkSkipFunc)spSkipID3Frame,
        
    (spChunkWriteHeaderFunc)spWriteID3FrameHeader,
    NULL,
    (spChunkWriteContentRestFunc)spWriteID3ContentRest,
        
    NULL,
};

spBool spInitID3Header(spID3Header *header, unsigned char version, unsigned char revision)
{
    if (header == NULL || !(version == 2 || version == 3 || version == 4)) {
        return SP_FALSE;
    }
    
    memset(header, 0, sizeof(spID3Header));
    header->version = version;
    header->revision = revision;

    return SP_TRUE;
}

spID3Frame *spAppendID3Frame(spID3Header *header, const char *type, spBool find_existing_frame)
{
    spID3Frame *frame = NULL;

    if (find_existing_frame == SP_FALSE || (frame = spFindID3FrameFromHeader(header, type)) == NULL) {
        if ((frame = spCreateID3Frame(header, type)) == NULL) {
            return NULL;
        }
    }

    return (spID3Frame *)frame;
}

void spPropagateID3FrameContentSize(spID3Frame *parent, spChunkSize size_diff)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    
    spPropagateChunkContentSize(&sp_id3_file_spec, (spChunk *)parent, size_diff);

    return;
}

spBool spSetID3FrameContentSize(spID3Frame *frame, spID3FrameSize size, spBool propagate_size)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spSetChunkContentSize(&sp_id3_file_spec, (spChunk *)frame, size, propagate_size);
}

spID3Frame *spCreateID3FrameWithCB(spID3Header *header, const char *type, spID3FrameUpdateFunc update_func, void *data)
{
    const char *vtype;
    spID3Frame *frame;
    
    if (header == NULL || strnone(type)) return NULL;
    
    if ((vtype = spGetID3CurrentVersionTypeId(header, type)) != NULL) {
        type = vtype;
    }
    
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }

    if ((frame = (spID3Frame *)spCreateChunk(&sp_id3_file_spec, (spChunk *)header, NULL, type,
                                             SP_TRUE, SP_TRUE, (spChunkUpdateFunc)update_func, data)) == NULL) {
        return NULL;
    }

    return frame;
}

spID3Frame *spCreateID3Frame(spID3Header *header, const char *type)
{
    return spCreateID3FrameWithCB(header, type, NULL, NULL);
}

spID3FrameSize spReadID3ChildFrame(spChunk *parent, long max_count, spID3FrameSize remain_size, long depth, int swap, FILE *fp)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spReadChildChunk(&sp_id3_file_spec, parent, max_count, remain_size, depth, swap, fp);
}

spID3FrameSize spReadID3Header(spID3Header *header, FILE *fp)
{
    spID3FrameSize nread;
    spID3FrameSize total_nread;

    memset(header, 0, sizeof(spID3Header));

    if ((nread = fread(header->header.type, 1, 3, fp)) != 3
        || !strneq(header->header.type, "ID3", 3)) {
        if (nread > 0) {
            spSeekFile(fp, -nread, SEEK_CUR);
        }
        spDebug(10, "spReadID3Header", "Can't read first 'ID3': %ld\n", (long)nread);
        return 0;
    }
    total_nread = nread;

    if ((nread = fread(&header->version, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read version: %ld\n", (long)nread);
        return 0;
    }
    total_nread += 1;

    if (header->version == 2 || header->version == 3 || header->version == 4) {
        spDebug(80, "spReadID3Header", "supported version: %d\n", header->version);
    } else {
        spDebug(10, "spReadID3Header", "unsupported version: %d\n", header->version);
        return 0;
    }
    
    if ((nread = fread(&header->revision, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read revision: %ld\n", (long)nread);
        return 0;
    }
    total_nread += 1;
    
    if ((nread = fread(&header->header.flag1, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read flag: %ld\n", (long)nread);
        return 0;
    }
    total_nread += 1;
    
    if ((nread = spReadID3SyncSafeLong32(&header->header.size, fp)) != 4) {
        spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
        return 0;
    }
    total_nread += nread;
    
    if ((nread = spReadID3ChildFrame((spChunk *)header, 0, header->header.size, 0, SP_AIFF_NEED_SWAP, fp)) <= 0) {
        spDebug(10, "spReadID3Header", "spReadID3ChildFrame failed: %ld\n", (long)nread);
        return 0;
    }
    total_nread += nread;

    spDebug(80, "spReadID3Header", "done: total_nread = %ld\n", (long)total_nread);
    
    return total_nread;
}

spChunkSize spWriteID3ChildFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spWriteChildChunk(&sp_id3_file_spec, frame, depth, SP_TRUE, rewrite_flag, swap, NULL, fp);
}

spID3FrameSize spWriteID3Header(spID3Header *header, FILE *fp)
{
    spID3FrameSize nwrite = 0;
    spID3FrameSize total_nwrite = 0;

    if (!strneq(header->header.type, "ID3", 3)
        || !(header->version == 2 || header->version == 3 || header->version == 4)) {
        spDebug(10, "spWriteID3Header", "not supported: version = %d\n", header->version);
        return 0;
    }

    if ((nwrite = fwrite(header->header.type, 1, 3, fp)) == 3) {
        total_nwrite = nwrite;
        if ((nwrite = fwrite(&header->version, 1, 1, fp)) == 1) {
            total_nwrite += nwrite;
            if ((nwrite = fwrite(&header->revision, 1, 1, fp)) == 1) {
                total_nwrite += nwrite;
                if ((nwrite = fwrite(&header->header.flag1, 1, 1, fp)) == 1) {
                    total_nwrite += nwrite;
                    if ((nwrite = spWriteID3SyncSafeLong32(header->header.size, fp)) == 4) {
                        total_nwrite += nwrite;
                    }
                }
            }
        }
    }

    spDebug(80, "spWriteID3Header", "write header: total_nwrite = %ld\n", (long)total_nwrite);
    
    if (total_nwrite != 10) {
        spDebug(10, "spWriteID3Header", "Can't write ID3 header: %ld\n", (long)nwrite);
        return 0;
    }

    if ((nwrite = spWriteID3ChildFrame((spChunk *)header, 0, SP_FALSE, SP_AIFF_NEED_SWAP, fp)) > 0) {
        total_nwrite += nwrite;
        spDebug(80, "spWriteID3Header", "done: total_nwrite = %ld\n", (long)total_nwrite);
        return total_nwrite;
    } else {
        spDebug(10, "spWriteID3Header", "spWriteID3ChildFrame failed: %ld\n", (long)nwrite);
        return 0;
    }
}

spBool spFreeID3Header(spID3Header *header)
{
    spDebug(80, "spFreeID3Header", "in\n");
    
    if (header != NULL) {
        return spFreeChunk(&sp_id3_file_spec, (spChunk *)header);
    } else {
        spDebug(80, "spFreeID3Header", "header == NULL\n");
        return SP_FALSE;
    }
}

spID3FrameSize spCopyID3ChildFrame(spChunk *dest_parent, spChunk *src_parent, long depth)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spCopyChildChunk(&sp_id3_file_spec, dest_parent, src_parent, depth);
}

spID3FrameSize spCopyID3Header(spID3Header *dest_header, spID3Header *src_header)
{
    spID3FrameSize ncopy;
    spID3FrameSize total_ncopy;

    memcpy(dest_header, src_header, sizeof(spID3Header));
    dest_header->child = NULL;

    if ((ncopy = spCopyID3ChildFrame((spChunk *)dest_header, (spChunk *)src_header, 0)) <= 0) {
        return ncopy;
    }
    total_ncopy = ncopy;

    spDebug(80, "spCopyID3Header", "done: total_ncopy = %ld\n", (long)total_ncopy);
    
    return total_ncopy;
}

spChunkInfoTable *spFindID3FrameInfoTable(spID3FrameHeader *parent_header, spID3FrameHeader *header)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }

    return spFindChunkInfoTable(sp_id3_frame_info_list, sp_id3_file_spec.num_list, (spChunkHeader *)parent_header, (spChunkHeader *)header);
}

spID3Frame *spFindID3FrameFromHeader(spID3Header *header, const char *type)
{
    const char *vtype;
    
    if ((vtype = spGetID3CurrentVersionTypeId(header, type)) == NULL) {
        spDebug(80, "spFindID3FrameFromHeader", "cannot find current version type: type = %s\n", type);
        return NULL;
    }
    
    return (spID3Frame *)spFindChildChunk((spChunk *)header, vtype, "ID3");
}

spID3Frame *spFindID3Frame(spID3Frame *frame, const char *type)
{
    const char *vtype;
    spID3Header *header;

    if ((header = spGetID3HeaderFromFrame(frame)) == NULL) {
        return NULL;
    }
    
    if ((vtype = spGetID3CurrentVersionTypeId(header, type)) == NULL) {
        return NULL;
    }
    
    return (spID3Frame *)spFindChunk((spChunk *)frame, type, "ID3");
}

const unsigned char *spGetID3TextFrameString(spID3TextFrame *tframe, unsigned char *encoding)
{
    spChunkInfoTable *info;
    unsigned char *string = NULL;

    if (tframe == NULL || tframe->parent == NULL) return NULL;

    if ((info = spFindID3FrameInfoTable(&tframe->parent->header, &tframe->header)) == NULL) {
        return NULL;
    }
    
    if (info->options & SP_ID3_CHUNK_INFO_OPTION_TEXT_MASK) {
        if (info->options & SP_ID3_CHUNK_INFO_OPTION_DESC_MASK) {
            spID3DescTextFrame *dtframe = (spID3DescTextFrame *)tframe;

            string = dtframe->actual_string;
        } else {
            string = tframe->string;
        }

        if (encoding != NULL) *encoding = tframe->encoding;
    }

    return string;
}

spBool spGetID3TextFrameStringToLocaleCode(spID3TextFrame *tframe, unsigned char *str_in_locale, long size)
{
    unsigned char encoding;
    const unsigned char *string;

    if ((string = spGetID3TextFrameString(tframe, &encoding)) == NULL) {
        return SP_FALSE;
    }
    spDebug(80, "spGetID3TextFrameStringToLocaleCode", "encoding = %d\n", encoding);

    if (encoding == 1/* UTF-16 */ || encoding == 2/* UTF-16BE */) {
        if (spUnicode16ToUTF8((unsigned short *)string, str_in_locale, size) <= 0) {
            return SP_FALSE;
        }
    } else {
        spStrCopy((char *)str_in_locale, size, (char *)string);
    }

    if (encoding != 0) {
        spConvertKanjiToLocaleCode(str_in_locale, size, SP_KANJI_CODE_UTF8);
    }

    spDebug(80, "spGetID3TextFrameStringToLocaleCode", "str_in_locale = %s\n", str_in_locale);
    
    return SP_TRUE;
}

spBool spGetID3TextFrameTrackNumber(spID3TextFrame *tframe, spLong32 *track, spLong32 *total)
{
    const unsigned char *string;
    char *p;
    char buf[SP_MAX_LINE];
    
    if ((string = spGetID3TextFrameString(tframe, NULL)) == NULL) return SP_FALSE;

    if (total != NULL) *total = 0;

    spStrCopy(buf, sizeof(buf), (char *)string);

    if ((p = spStrChr(buf, '/')) != NULL) {
        if (total != NULL) *total = atol(p + 1);
        *p = NUL;
    }

    if (track != NULL) *track = atol(buf);

    return SP_TRUE;
}

spBool spGetID3TextFrameNumber(spID3TextFrame *tframe, spLong32 *number)
{
    const unsigned char *string;
    char buf[SP_MAX_LINE];
    
    if ((string = spGetID3TextFrameString(tframe, NULL)) == NULL) return SP_FALSE;

    spStrCopy(buf, sizeof(buf), (const char *)string);

    if (number != NULL) *number = atol(buf);
    
    return SP_TRUE;
}

const unsigned char *spFindID3TextFrameString(spID3Header *header, const char *type, unsigned char *encoding)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, type)) == NULL) {
        return NULL;
    }

    return spGetID3TextFrameString(tframe, encoding);
}

spBool spFindID3TextFrameStringToLocaleCode(spID3Header *header, const char *type, unsigned char *str_in_locale, long size)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, type)) == NULL) {
        spDebug(80, "spFindID3TextFrameTrackNumber", "cannot find frame: type = %s\n", type);
        return SP_FALSE;
    }

    return spGetID3TextFrameStringToLocaleCode(tframe, str_in_locale, size);
}

spBool spFindID3TextFrameTrackNumber(spID3Header *header, spLong32 *track, spLong32 *total)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_TRACK)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameTrackNumber(tframe, track, total);
}

spBool spFindID3TextFrameDiscNumber(spID3Header *header, spLong32 *disc, spLong32 *total)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_DISC)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameTrackNumber(tframe, disc, total);
}

spBool spFindID3TextFrameTempo(spID3Header *header, spLong32 *tempo)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_TEMPO)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameNumber(tframe, tempo);
}

spBool spSetID3TextFrameString(spID3TextFrame *tframe, unsigned char encoding, const unsigned char *str, long len)
{
    if (tframe == NULL || str == NULL || len <= 0) return SP_FALSE;

    if (tframe->string != NULL) xfree(tframe->string);
    
    tframe->encoding = encoding;
    
    tframe->string = xalloc(len + 2, unsigned char);
    memcpy(tframe->string, str, len);
    tframe->string[len] = 0;
    tframe->string[len + 1] = 0;

    spSetID3FrameContentSize((spID3Frame *)tframe, len + 1, SP_TRUE);

    return SP_TRUE;
}

spBool spSetID3TextFrameStringFromLocaleCode(spID3TextFrame *tframe, unsigned char encoding, const unsigned char *str_in_locale)
{
    long size;
    unsigned char *ptr;
    spKanjiCode icode;
    unsigned char *utf8str = NULL;
    unsigned short *utf16str = NULL;
    
    if (tframe == NULL || str_in_locale == NULL) return SP_FALSE;
    
    size = (long)(strlen((char *)str_in_locale) + 1) * 2;
    utf8str = xalloc(size, unsigned char);
    spStrCopy((char *)utf8str, size, (char *)str_in_locale);

    icode = spConvertKanjiFromLocaleCode(utf8str, size, SP_KANJI_CODE_UTF8);
    spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "encoding = %d, icode = %d\n", encoding, icode);

    if (encoding == 1/* UTF-16 */ || encoding == 2/* UTF-16BE */) {
        size = MAX(size, (long)(strlen((char *)utf8str) + 1));
        utf16str = xalloc(size, unsigned short);
        
        spUTF8ToUnicode16(utf8str, utf16str, size * 2);
        ptr = (unsigned char *)utf16str;
    } else {
        ptr = utf8str;
    }
    
    size = (long)strlen((char *)ptr);
    
    spSetID3TextFrameString(tframe, encoding, ptr, size);

    xfree(utf8str);
    if (utf16str != NULL) {
        xfree(utf16str);
    }

    return SP_TRUE;
}

spID3TextFrame *spAppendID3TextFrameStringFromLocaleCode(spID3Header *header, const char *type, spBool find_existing_frame, unsigned char encoding, const unsigned char *str_in_locale)
{
    spID3TextFrame *tframe = NULL;
    
    if (header == NULL || (encoding == 3/* UTF-8 */ && header->version < 4) || str_in_locale == NULL) return NULL;

    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, find_existing_frame)) == NULL) {
        return NULL;
    }
    
    spSetID3TextFrameStringFromLocaleCode(tframe, encoding, str_in_locale);

    return tframe;
}

spID3TextFrame *spAppendID3TextFrameString(spID3Header *header, const char *type, spBool find_existing_frame,
                                           unsigned char encoding, const unsigned char *str, long len)
{
    spID3TextFrame *tframe = NULL;
    
    if (header == NULL || (encoding == 3/* UTF-8 */ && header->version < 4) || str == NULL) return NULL;

    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, find_existing_frame)) == NULL) {
        return NULL;
    }
    
    spSetID3TextFrameString(tframe, encoding, str, len);

    return tframe;
}

static spID3TextFrame *appendID3TextFrameTrackOrDiscNumber(spID3Header *header, spLong32 number, spLong32 total, spBool disc_flag)
{
    char *type;
    char buf[SP_MAX_LINE];
    spID3TextFrame *tframe = NULL;

    if (disc_flag) {
        type = SP_ID3_FRAME_TYPE_DISC;
    } else {
        type = SP_ID3_FRAME_TYPE_TRACK;
    }
    
    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, SP_TRUE)) == NULL) {
        return NULL;
    }

    if (total > 0) {
        sprintf(buf, "%d/%d", (int)number, (int)total);
    } else {
        sprintf(buf, "%d", (int)number);
    }
    
    spSetID3TextFrameString(tframe, 0, (const unsigned char *)buf, (long)strlen(buf));

    return tframe;
}

spID3TextFrame *spAppendID3TextFrameTrackNumber(spID3Header *header, spLong32 number, spLong32 total)
{
    return appendID3TextFrameTrackOrDiscNumber(header, number, total, SP_FALSE);
}

spID3TextFrame *spAppendID3TextFrameDiscNumber(spID3Header *header, spLong32 number, spLong32 total)
{
    return appendID3TextFrameTrackOrDiscNumber(header, number, total, SP_TRUE);
}

spID3TextFrame *spAppendID3TextFrameTempo(spID3Header *header, spLong32 tempo)
{
    char buf[SP_MAX_LINE];
    spID3TextFrame *tframe = NULL;

    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, SP_ID3_FRAME_TYPE_TEMPO, SP_TRUE)) == NULL) {
        return NULL;
    }

    sprintf(buf, "%d", (int)tempo);
    
    spSetID3TextFrameString(tframe, 0, (unsigned char *)buf, (long)strlen(buf));

    return tframe;
}

spBool spParseID3SongInfo(spID3Header *header, spSongInfoV2 *song_info)
{
    if (header == NULL || song_info == NULL) return SP_FALSE;

    spInitSongInfoV2(song_info);

    if (header->version == 3) {
        song_info->info_mask |= SP_SONG_ID3V2_3_MASK;
    } else if (header->version == 2) {
        song_info->info_mask |= SP_SONG_ID3V2_2_MASK;
    } else {
        song_info->info_mask |= SP_SONG_ID3V2_4_MASK;
    }

    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_TITLE, (unsigned char *)song_info->title, sizeof(song_info->title))) {
        song_info->info_mask |= SP_SONG_TITLE_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_ARTIST, (unsigned char *)song_info->artist, sizeof(song_info->artist))) {
        song_info->info_mask |= SP_SONG_ARTIST_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_ALBUM, (unsigned char *)song_info->album, sizeof(song_info->album))) {
        song_info->info_mask |= SP_SONG_ALBUM_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_GENRE, (unsigned char *)song_info->genre, sizeof(song_info->genre))) {
        song_info->info_mask |= SP_SONG_GENRE_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_ALBUM_ARTIST, (unsigned char *)song_info->album_artist, sizeof(song_info->album_artist))) {
        song_info->info_mask |= SP_SONG_ALBUM_ARTIST_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_RELEASE, (unsigned char *)song_info->release, sizeof(song_info->release))) {
        song_info->info_mask |= SP_SONG_RELEASE_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_COPYRIGHT, (unsigned char *)song_info->copyright, sizeof(song_info->copyright))) {
        song_info->info_mask |= SP_SONG_COPYRIGHT_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_ISRC, (unsigned char *)song_info->isrc, sizeof(song_info->isrc))) {
        song_info->info_mask |= SP_SONG_ISRC_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_COMPOSER, (unsigned char *)song_info->composer, sizeof(song_info->composer))) {
        song_info->info_mask |= SP_SONG_COMPOSER_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_LYRICIST, (unsigned char *)song_info->lyricist, sizeof(song_info->lyricist))) {
        song_info->info_mask |= SP_SONG_LYRICIST_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_SOFTWARE, (unsigned char *)song_info->software, sizeof(song_info->software))) {
        song_info->info_mask |= SP_SONG_SOFTWARE_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_PUBLISHER, (unsigned char *)song_info->source, sizeof(song_info->source))) {
        song_info->info_mask |= SP_SONG_SOURCE_MASK;
    }
    if (spFindID3TextFrameStringToLocaleCode(header, SP_ID3_FRAME_TYPE_COMMENT, (unsigned char *)song_info->comment, sizeof(song_info->comment))) {
        song_info->info_mask |= SP_SONG_COMMENT_MASK;
    }

    if (spFindID3TextFrameTrackNumber(header, &song_info->track, &song_info->track_total)) {
        if (song_info->track > 0) {
            song_info->info_mask |= SP_SONG_TRACK_MASK;
        }
        if (song_info->track_total > 0) {
            song_info->info_mask |= SP_SONG_TRACK_TOTAL_MASK;
        }
    }

    if (spFindID3TextFrameDiscNumber(header, &song_info->disc, &song_info->disc_total)) {
        if (song_info->disc > 0) {
            song_info->info_mask |= SP_SONG_DISC_MASK;
        }
        if (song_info->disc_total > 0) {
            song_info->info_mask |= SP_SONG_DISC_TOTAL_MASK;
        }
    }

    if (spFindID3TextFrameTempo(header, &song_info->tempo)) {
        song_info->info_mask |= SP_SONG_TEMPO_MASK;
    }
    
    if (song_info->info_mask) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spAppendID3SongInfo(spID3Header *header, spSongInfoV2 *song_info)
{
    spLong32 number, total;
    
    if (header == NULL || song_info == NULL || song_info->info_mask == 0) return SP_FALSE;

    spDebug(10, "spAppendID3SongInfo", "song_info->info_mask = %lx\n", song_info->info_mask);

    if (song_info->info_mask & SP_SONG_ID3V2_4_MASK) {
        header->version = 4;
    } else if (song_info->info_mask & SP_SONG_ID3V2_3_MASK) {
        header->version = 3;
    } else if (song_info->info_mask & SP_SONG_ID3V2_2_MASK) {
        header->version = 2;
    } else {
        return SP_FALSE;
    }
    
    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_TITLE, SP_TRUE, 1, (unsigned char *)song_info->title);
    }
    if (song_info->info_mask & SP_SONG_ARTIST_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_ARTIST, SP_TRUE, 1, (unsigned char *)song_info->artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_ARTIST_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_ALBUM_ARTIST, SP_TRUE, 1, (unsigned char *)song_info->album_artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_ALBUM, SP_TRUE, 1, (unsigned char *)song_info->album);
    }
    if (song_info->info_mask & SP_SONG_ISRC_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_ISRC, SP_TRUE, 1, (unsigned char *)song_info->isrc);
    }
    if (song_info->info_mask & SP_SONG_COMPOSER_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_COMPOSER, SP_TRUE, 1, (unsigned char *)song_info->composer);
    }
    if (song_info->info_mask & SP_SONG_LYRICIST_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_LYRICIST, SP_TRUE, 1, (unsigned char *)song_info->lyricist);
    }
    if (song_info->info_mask & SP_SONG_GENRE_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_GENRE, SP_TRUE, 1, (unsigned char *)song_info->genre);
    }
    if (song_info->info_mask & SP_SONG_RELEASE_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_RELEASE, SP_TRUE, 1, (unsigned char *)song_info->release);
    }
    if (song_info->info_mask & SP_SONG_SOFTWARE_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_SOFTWARE, SP_TRUE, 1, (unsigned char *)song_info->software);
    }
    if (song_info->info_mask & SP_SONG_COPYRIGHT_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_COPYRIGHT, SP_TRUE, 1, (unsigned char *)song_info->copyright);
    }
    if (song_info->info_mask & SP_SONG_SOURCE_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_PUBLISHER, SP_TRUE, 1, (unsigned char *)song_info->source);
    }
    if (song_info->info_mask & SP_SONG_COMMENT_MASK) {
        spAppendID3TextFrameStringFromLocaleCode(header, SP_ID3_FRAME_TYPE_COMMENT, SP_TRUE, 1, (unsigned char *)song_info->comment);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_TRACK_MASK) {
        number = song_info->track;
        if (song_info->info_mask & SP_SONG_TRACK_TOTAL_MASK) {
            total = song_info->track_total;
        }
        spAppendID3TextFrameTrackNumber(header, number, total);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_DISC_MASK) {
        number = song_info->disc;
        if (song_info->info_mask & SP_SONG_DISC_TOTAL_MASK) {
            total = song_info->disc_total;
        }
        spAppendID3TextFrameDiscNumber(header, number, total);
    }

    if (song_info->info_mask & SP_SONG_TEMPO_MASK && song_info->tempo > 0) {
        spAppendID3TextFrameTempo(header, song_info->tempo);
    }

    return SP_TRUE;
}
